import pandas as pd
import pathlib
import matplotlib.pyplot as plt
import seaborn as sns
from matplotlib.patches import Rectangle

from ..utils import helper_functions, check_functions


def summarize_results_per_phenotype_and_datasplit(results_directory_genotype_level: str, eval_metric: str = None):
    """
    Summarize the results for each phenotype and datasplit for all models and save in a file.

    The following files will be created:

        - at phenotype-folder level within results directories:

            - Detailed_results_summary_*PHENOTYPE*DATASPLIT-PATTERN*.xlsx: .xlsx-file containing detailed results for each phenotype and datasplit-maf pattern (e.g. with all runtime results etc.)
            - Results_summary**PHENOTYPE*DATASPLIT-PATTERN*.csv: .csv-file containing an overview of the performance of each model for a phenotype and datasplit-maf pattern combination

        - at genotype-folder level within results directories (the one that was specified):

            - Results_summary_all_phenotypes*DATASPLIT-PATTERN*.xlsx: .xlsx-file containing an overview of the performance of each model on each phenotype used for this genotype matrix with the specified datasplit-maf pattern
            - Results_summary_all_phenotypes*EVALMETRIC*DATASPLIT-PATTERN*.csv: only overview sheet of Results_summary*DATASPLIT-PATTERN*.xlsx

    :param results_directory_genotype_level: results directory at the level of the name of the genotype matrix
    :param eval_metric: eval metric to use for summary. Options (default given first):
            Regression: explained_variance, r2_score, rmse, mse.
            Classification: mcc, f1_score, accuracy, precision, recall.

    """
    results_directory_genotype_level = pathlib.Path(results_directory_genotype_level)

    # Check user inputs
    print("Checking user inputs")
    if not check_functions.check_exist_directories(list_of_dirs=[results_directory_genotype_level]):
        raise Exception("See output above. Problems with specified directories")
    if not results_directory_genotype_level.parts[-2] == 'results':
        raise Exception("Problems with specified directory: " + str(results_directory_genotype_level) +
                        "\n Make sure the results directory is at the level fo the genotype matrix name.")
    if eval_metric is not None:
        if eval_metric not in ["explained_variance", "r2_score", "rmse", "mse",
                               "mcc", "f1_score", "accuracy", "precision", "recall"]:
            raise Exception("Invalid eval metric: " + eval_metric)

    for phenotype_matrix in \
            sorted(helper_functions.get_all_subdirectories_non_recursive(results_directory_genotype_level)):
        results_directory_phenotype_matrix_level = results_directory_genotype_level.joinpath(phenotype_matrix)
        for phenotype_folder in \
                helper_functions.get_all_subdirectories_non_recursive(results_directory_phenotype_matrix_level):
            phenotype = phenotype_folder.parts[-1]
            print('++++++++++++++ PHENOTYPE ' + phenotype + ' ++++++++++++++')
            subdirs = [fullpath.parts[-1]
                       for fullpath in helper_functions.get_all_subdirectories_non_recursive(phenotype_folder)]
            datasplit_maf_patterns = set(['_'.join(path.split('_')[:3]) for path in subdirs])
            for pattern in list(datasplit_maf_patterns):
                writer = pd.ExcelWriter(
                    phenotype_folder.joinpath('Detailed_results_summary_' + phenotype + '_' + pattern + '.xlsx'),
                    engine='xlsxwriter'
                )
                overview_df = None
                print('----- Datasplit pattern ' + pattern + ' -----')
                print('Got results for ' + str(len([model for path in phenotype_folder.glob(pattern + '*')
                                                    for model in path.parts[-1].split('_')[3].split('+')])) +
                      ' models.')
                for path in phenotype_folder.glob(pattern + '*'):
                    models = path.parts[-1].split('_')[3].split('+')
                    for current_model in models:
                        print('### Results for ' + current_model + ' ###')
                        try:
                            results_file = list(path.glob('Results*.csv'))[0]
                            results = pd.read_csv(results_file)
                            results = results.loc[:, [current_model in col for col in results.columns]]
                            if 'nested' in pattern:
                                eval_dict_std = result_string_to_dictionary(
                                    result_string=results[current_model + '___eval_metrics'].iloc[-1])
                                eval_dict = result_string_to_dictionary(
                                    result_string=results[current_model + '___eval_metrics'].iloc[-2]
                                )
                                runtime_dict_std = result_string_to_dictionary(
                                    result_string=results[current_model + '___runtime_metrics'].iloc[-1]
                                )
                                runtime_dict = result_string_to_dictionary(
                                    result_string=results[current_model + '___runtime_metrics'].iloc[-2]
                                )
                            else:
                                eval_dict = result_string_to_dictionary(
                                    result_string=results[current_model + '___eval_metrics'][0]
                                )
                                runtime_dict = result_string_to_dictionary(
                                    result_string=results[current_model + '___runtime_metrics'][0]
                                )
                            results.to_excel(writer, sheet_name=current_model + '_results', index=False)
                            eval_dict = {str(key) + '_mean': val for key, val in eval_dict.items()}
                            runtime_dict = {str(key) + '_mean': val for key, val in runtime_dict.items()}
                            new_row = {'model': current_model}
                            new_row.update(eval_dict)
                            new_row.update(runtime_dict)
                            if 'nested' in pattern:
                                eval_dict_std = {str(key) + '_std': val for key, val in eval_dict_std.items()}
                                runtime_dict_std = {str(key) + '_std': val for key, val in runtime_dict_std.items()}
                                new_row.update(eval_dict_std)
                                new_row.update(runtime_dict_std)
                            if overview_df is None:
                                overview_df = pd.DataFrame(new_row, index=[0])
                            else:
                                overview_df = pd.concat([overview_df, pd.DataFrame(new_row, index=[0])],
                                                        ignore_index=True)
                        except Exception as exc:
                            print('No Results File')
                            continue
                        if 'nested' in pattern:
                            for outerfold_path in path.glob('outerfold*'):
                                runtime_file = pd.read_csv(
                                    outerfold_path.joinpath(current_model, current_model + '_runtime_overview.csv')
                                )
                                runtime_file.to_excel(
                                    writer, sheet_name=current_model + '_of' + outerfold_path.parts[-1].split('_')[-1]
                                                       + '_runtime',
                                    index=False
                                )
                        else:
                            runtime_file = \
                                pd.read_csv(path.joinpath(current_model, current_model + '_runtime_overview.csv'))
                            runtime_file.to_excel(writer, sheet_name=current_model + '_runtime', index=False)
                if overview_df is None:
                    continue
                overview_df.to_excel(writer, sheet_name='Overview_results', index=False)
                overview_df.to_csv(phenotype_folder.joinpath('Results_summary_' + phenotype + '_' + pattern + '.csv'))
                writer.sheets['Overview_results'].activate()
                writer.save()
    for pattern in datasplit_maf_patterns:
        overview_sheet = pd.DataFrame(
            columns=helper_functions.get_list_of_implemented_models()
        )
        writer = pd.ExcelWriter(
            results_directory_genotype_level.joinpath('Results_summary_all_phenotypes_' + pattern + '.xlsx'),
            engine='xlsxwriter'
        )
        paths = [path for path in list(results_directory_genotype_level.rglob('Results_summary*' + pattern + '*.csv'))
                 if 'all_phenotypes' not in str(path)]
        overview_sheet['phenotype'] = [path.parts[-2] for path in paths]
        overview_sheet.set_index('phenotype', drop=True, inplace=True)
        for results_summary_path in paths:
            results_summary = pd.read_csv(results_summary_path)
            results_summary.to_excel(
                writer, sheet_name=results_summary_path.parts[-2],
                index=False
            )
            phenotype = results_summary_path.parts[-2]
            if eval_metric is None:
                eval_metric = 'test_explained_variance' \
                    if any(['test_explained_variance' in col for col in results_summary.columns]) else 'test_mcc'
            else:
                if 'test' not in eval_metric:
                    eval_metric = 'test_' + eval_metric
            if 'nested' in pattern:
                for row in results_summary.iterrows():
                    overview_sheet.at[phenotype, row[1]['model']] = "{:.3f} +- {:.3f}".format(
                        row[1][eval_metric + '_mean'], row[1][eval_metric + '_std'])
            else:
                for row in results_summary.iterrows():
                    overview_sheet.at[phenotype, row[1]['model']] = "{:.3f}".format(row[1][eval_metric + '_mean'])
        overview_sheet.dropna(axis=1, inplace=True, how='all')  # drop model column if all results are missing
        overview_sheet = overview_sheet.sort_index()
        overview_sheet.to_excel(writer, sheet_name='Overview')
        overview_sheet.to_csv(
            results_directory_genotype_level.joinpath(
                'Results_summary_all_phenotypes_' + eval_metric[5:] + '_' + pattern + '.csv'
            )
        )
        writer.sheets['Overview'].activate()
        writer.save()


def result_string_to_dictionary(result_string: str) -> dict:
    """
    Convert result string saved in a .csv file to a dictionary

    :param result_string: string from .csv file

    :return: dictionary with info from .csv file
    """
    key_value_strings = result_string.split('\\')[0][2:-2].replace('\'', '').split(',')
    dict_result = {}
    for key_value_string in key_value_strings:
        key = key_value_string.split(':')[0].strip()
        value = key_value_string.split(':')[1].strip()
        try:
            value = float(value)
            value = int(value) if value == int(value) else value
        except:
            value = value
        dict_result[key] = value
    return dict_result


def plot_heatmap_results(path_to_results_summary_csv: str, save_dir: str):
    """
    Generate a heatmap based on the results summary .csv file

    :param path_to_results_summary_csv: path to the results summary .csv file
    :param save_dir: directory to save the plots
    """
    path_to_results_summary_csv = pathlib.Path(path_to_results_summary_csv)
    save_dir = pathlib.Path(save_dir) if save_dir is not None else path_to_results_summary_csv.parents[0]
    fig, ax = plt.subplots(figsize=(12, 6))

    results_overview = pd.read_csv(path_to_results_summary_csv)
    results_overview.set_index("phenotype", inplace=True)
    models = [mod for mod in results_overview.columns if mod != 'phenotype']

    if 'nested' in path_to_results_summary_csv.parts[-1]:
        types = ['mean', 'std']
    else:
        types = ['mean']
    plot_data_full = pd.DataFrame(columns=[mod + '_' + type for mod in models for type in types] + ['phenotype'])
    plot_data_full['phenotype'] = results_overview.index
    plot_data_full.set_index("phenotype", inplace=True)
    for row in results_overview.iterrows():
        for model in models:
            result = row[1][model]
            if 'std' in types:
                plot_data_full.at[row[0], model + '_std'] = float(result.split('+-')[1])
                plot_data_full.at[row[0], model + '_mean'] = float(result.split('+-')[0])
            else:
                plot_data_full.at[row[0], model + '_mean'] = float(result)
    plot_data_mean = plot_data_full.filter(regex='mean').astype(float)
    row_max = plot_data_mean.idxmax(axis=1)
    sns.heatmap(data=plot_data_mean, cmap="Spectral", cbar_kws={"shrink": .75},
                annot=results_overview, fmt='', linewidths=1.5, linecolor='white', cbar=True, annot_kws={"size": 12})
    ax.set_xticklabels(results_overview.columns, rotation=0)
    ax.set_yticklabels(plot_data_full.index, rotation=0)
    ax.tick_params(top=False,
                   bottom=False,
                   left=False,
                   right=False,
                   labelleft=True,
                   labelbottom=True)
    for row, index in enumerate(plot_data_mean.index):
        position = results_overview.columns.get_loc(row_max[index].split('_')[0])
        ax.add_patch(Rectangle((position, row), 1, 1, fill=False, edgecolor='0', lw=1.5))

    fig.tight_layout()
    plt.savefig(save_dir.joinpath('heatmap_' + path_to_results_summary_csv.parts[-1].split('.')[0] + '.pdf'),
                bbox_inches='tight', dpi=600)



import pandas as pd
import numpy as np
import pathlib
import matplotlib.pyplot as plt

from ..utils import helper_functions


def gather_sim_configs(sim_config_dir: pathlib.Path, save_dir: pathlib.Path):
    """
    Collect the information on the simulation configurations for all within the specified directory

    :param sim_config_dir: directory which contains the sim config files
    :param save_dir: directory to save the collected sim config info
    """
    all_sim_configs = pd.DataFrame(columns=['sim_id', 'type', 'snp_id', 'beta'])
    config_files = sim_config_dir.expanduser().glob('simulation_config_*')
    for file in config_files:
        sim_id = file.with_suffix('').name.split('_')[-1]
        background_file = pd.read_csv(sim_config_dir.joinpath('background_' + sim_id + '.csv'))
        betas_background_file = pd.read_csv(sim_config_dir.joinpath('betas_background_' + sim_id + '.csv'))
        sim_config_file = pd.read_csv(sim_config_dir.joinpath('simulation_config_' + sim_id + '.csv'))
        df_to_append = pd.DataFrame(columns=['sim_id', 'type', 'snp_id', 'beta'])
        if '-' in sim_id:
            sim_id_parts = sim_id.split('-')
            sim_ids = np.arange(int(sim_id_parts[0]), int(sim_id_parts[-1]) + 1)
        else:
            sim_ids = [int(sim_id)]
        for number in sim_ids:
            column = 'sim' + str(number)
            df_to_append['snp_id'] = background_file[column]
            df_to_append['type'] = 'background'
            df_to_append['sim_id'] = column
            df_to_append['beta'] = betas_background_file[column]
            sim_conf = sim_config_file[sim_config_file['simulation'] == number]
            for index, causal_marker in enumerate(
                    sim_conf['causal_marker'][sim_conf.index[0]][2:-2].replace('\n', '').split('\' \'')):
                row_id = df_to_append.shape[0]
                df_to_append.at[row_id, 'sim_id'] = column
                df_to_append.at[row_id, 'snp_id'] = causal_marker
                df_to_append.at[row_id, 'type'] = 'causal'
                df_to_append.at[row_id, 'beta'] = float(sim_conf['causal_beta'][sim_conf.index[0]][1:-1].split(',')[index])
            all_sim_configs = pd.concat([all_sim_configs, df_to_append], ignore_index=True)
    all_sim_configs.to_csv(save_dir.joinpath('Sim_configs_gathered_' + sim_config_dir.parts[-2] + '.csv'),
                           index=False)


def gather_feature_importances(results_dir: pathlib.Path, save_dir: pathlib.Path, datasplit_maf_pattern: str):
    """
    Collect the information on the feature importances for all models within the specified directory and for the specified datasplit maf pattern

    :param results_dir: results directory at the level of the name of the genotype matrix
    :param save_dir: directory to save the collected info
    :param datasplit_maf_pattern: datasplit maf pattern to search on
    """

    all_feat_imps = pd.DataFrame(columns=['sim_id', 'model', 'snp_id', 'feat_imp'])
    for phenotype_matrix in helper_functions.get_all_subdirectories_non_recursive(results_dir):
        if 'Simulation' not in phenotype_matrix.name:
            continue
        for phenotype_folder in \
                helper_functions.get_all_subdirectories_non_recursive(phenotype_matrix):
            phenotype = phenotype_folder.parts[-1]
            print('++++++++++++++ PHENOTYPE ' + phenotype + ' ++++++++++++++')
            for path in phenotype_folder.glob(datasplit_maf_pattern + '*'):
                if len(list(path.glob('Results_over*.csv'))) == 0:
                    continue
                models = path.parts[-1].split('_')[3].split('+')
                print('working on ' + str(path))
                for current_model in models:
                    if path.joinpath('outerfold_0', current_model, 'final_model_feature_importances.csv').is_file() or \
                            path.joinpath(current_model, 'final_model_feature_importances.csv').is_file():
                        feat_imps_to_append = None
                        print('Model: ' + current_model)
                        iterations = int(datasplit_maf_pattern.split('_')[1].split('-')[0]) \
                            if 'nested' in datasplit_maf_pattern else 1
                        for outerfold_nr in range(iterations):
                            current_directory = path.joinpath('outerfold_' + str(outerfold_nr), current_model) \
                                if 'nested' in datasplit_maf_pattern else path.joinpath(current_model)
                            try:
                                feat_imp = \
                                    pd.read_csv(current_directory.joinpath('final_model_feature_importances.csv'))
                                feat_imp.rename(columns={'snp_ids_0': 'snp_id'}, inplace=True)
                                feat_imp.rename(columns={'snp_ids_standard': 'snp_id'}, inplace=True)
                                if feat_imps_to_append is None:
                                    feat_imps_to_append = feat_imp.copy()
                                else:
                                    feat_imps_to_append = feat_imps_to_append.join(feat_imp.set_index('snp_id'),
                                                                                   on=['snp_id'],
                                                                                   rsuffix=current_directory.parts[-2],
                                                                                   how='outer')
                            except Exception as exc:
                                print(exc)
                        if feat_imps_to_append is not None:
                            cols_to_drop = feat_imps_to_append.columns[1:]
                            feat_imps_to_append['feat_imp'] = feat_imps_to_append.iloc[:, 1:].fillna(0).mean(axis=1)
                            feat_imps_to_append['sim_id'] = phenotype
                            feat_imps_to_append['model'] = current_model
                            feat_imps_to_append.drop(cols_to_drop, axis=1, inplace=True)
                            all_feat_imps = pd.concat([all_feat_imps, feat_imps_to_append], ignore_index=True)
    all_feat_imps.to_csv(
        save_dir.joinpath('Feature_importances_gathered_' + results_dir.parts[-1] + datasplit_maf_pattern + '.csv'),
        index=False
    )


def get_statistics_featimps_vs_simulation(all_sim_configs: pd.DataFrame, all_feat_imps: pd. DataFrame,
                                          min_perc_threshold: float = 0.01) -> pd.DataFrame:
    """
    Get statistics on feature importances compared to effect sizes on synthetic data, e.g. on how many background SNPs were detected

    :param all_sim_configs: simulation configs to consider
    :param all_feat_imps: feature importances to consider
    :param min_perc_threshold: threshold for minimum feature importance in relation to maximum feature importance for a specific model

    :return: statistics for a comparison between feature importances and effect sizes in a DataFrame
    """
    stats = None
    for sim_id in set(all_sim_configs.sim_id):
        config_simid = all_sim_configs[all_sim_configs.sim_id == sim_id]
        feat_imp_simid = all_feat_imps[all_feat_imps.sim_id == sim_id]
        for model in set(feat_imp_simid.model):
            new_row = {}
            new_row['sim_id'] = sim_id
            new_row['model'] = model
            feat_imp_model = feat_imp_simid[feat_imp_simid.model == model]
            feat_imp_model_perc_threshold = feat_imp_model[
                feat_imp_model['feat_imp'].abs() > min_perc_threshold * feat_imp_model['feat_imp'].abs().max()]
            non_zero_featimps = feat_imp_model[feat_imp_model.feat_imp != 0.0]
            non_zero_featimps_perc_threshold = feat_imp_model_perc_threshold[
                feat_imp_model_perc_threshold.feat_imp != 0.0]
            new_row['#features_with_imp'] = non_zero_featimps.shape[0]
            new_row['#features_with_imp_minpercmax_' + str(min_perc_threshold)] = \
                non_zero_featimps_perc_threshold.shape[0]
            background = config_simid[config_simid.type == 'background']
            causal = config_simid[config_simid.type == 'causal']
            new_row['#detected_background_snps'] = len(
                set(non_zero_featimps.snp_id).intersection(set(background.snp_id)))
            new_row['#detected_background_snps_minpercmax_' + str(min_perc_threshold)] = len(
                set(non_zero_featimps_perc_threshold.snp_id).intersection(set(background.snp_id)))
            new_row['#background_snps'] = len(set(background.snp_id))
            new_row['ratio_background_detected_by_feat_imp'] = \
                round(new_row['#detected_background_snps'] / new_row['#features_with_imp'], 2) \
                    if new_row[ '#features_with_imp'] != 0 else np.nan
            new_row['ratio_background_detected'] = round(
                new_row['#detected_background_snps'] / new_row['#background_snps'], 2)
            new_row['ratio_background_detected_by_feat_imp_minpercmax_' + str(min_perc_threshold)] = \
                round(new_row['#detected_background_snps_minpercmax_' + str(min_perc_threshold)] /
                      new_row['#features_with_imp_minpercmax_' + str(min_perc_threshold)], 2) \
                    if new_row['#features_with_imp_minpercmax_' + str(min_perc_threshold)] != 0 else np.nan
            new_row['ratio_background_detected_minpercmax_' + str(min_perc_threshold)] = round(
                new_row['#detected_background_snps_minpercmax_' + str(min_perc_threshold)] / new_row['#background_snps'], 2)
            new_row['#detected_causal_snps'] = len(set(non_zero_featimps.snp_id).intersection(set(causal.snp_id)))
            new_row['#causal_snps'] = len(set(causal.snp_id))
            new_row['ratio_causal_detected'] = round(new_row['#detected_causal_snps'] / new_row['#causal_snps'], 2)
            new_row['causal_snps_detected'] = 'Yes' if new_row['ratio_causal_detected'] == 1 else 'No'
            featimp_sorted = non_zero_featimps.sort_values(by='feat_imp', ascending=False).reset_index(drop=True)
            new_row['featimp_rank_causal_snps'] = str(list(
                featimp_sorted[featimp_sorted.snp_id.isin(set(causal.snp_id))].index + 1)) \
                if new_row['#detected_causal_snps'] != 0 else None
            if stats is None:
                stats = pd.DataFrame(new_row, index=[0])
            else:
                stats = pd.concat([stats, pd.DataFrame(new_row, index=[0])], ignore_index=True)
    stats = stats.sort_values(by=['sim_id', 'model'])
    return stats


def generate_scatterplots_featimps_vs_simulation(all_feat_imps: pd.DataFrame, all_sim_configs: pd.DataFrame,
                                                 save_dir: pathlib.Path, datasplit_maf_pattern: str):
    """
    Generate scatterplots based on feature importances and effect sizes on synthetic data. One plot containing all models for the specified datasplit maf pattern as well as single plots for each model will be generated and saved.

    :param all_sim_configs: simulation configs to consider
    :param all_feat_imps: feature importances to consider
    :param save_dir: directory to save the plots
    :param datasplit_maf_pattern: datasplit maf pattern to search on
    """
    pd.options.mode.chained_assignment = None
    sim_ids = sorted(set(all_feat_imps['sim_id']))
    sim_infos = all_sim_configs[all_sim_configs['sim_id'].isin(sim_ids)]
    beta_max = sim_infos[sim_infos['beta'] != 0]['beta'].astype(float).abs().max()
    models_total = set(all_feat_imps['model'])
    iter_models = []
    for single_model in models_total:
        iter_models.append([single_model])
    iter_models.append([single_model for single_model in models_total])
    for index_itermodels, models in enumerate(iter_models):
        full_plot = True if index_itermodels == len(iter_models) - 1 else False
        print(models)
        rows = int(np.ceil(len(set(sim_ids)) / 3))
        fig = plt.figure(figsize=(16, rows*3))
        for plot_nr, sim_id in enumerate(sim_ids):
            ax = fig.add_subplot(rows, min(3, len(set(sim_ids))), plot_nr + 1)
            sim_info_for_id = sim_infos[sim_infos['sim_id'] == sim_id]
            feat_info_for_id = all_feat_imps[all_feat_imps['sim_id'] == sim_id]
            for index_model, model in enumerate(models):
                color_index = index_itermodels if index_itermodels != len(iter_models) - 1 else index_model
                feat_info_for_model = feat_info_for_id[feat_info_for_id['model'] == model]
                plot_info = sim_info_for_id.iloc[:, 1:].join(feat_info_for_model.iloc[:, 2:].set_index('snp_id'),
                                                             on=['snp_id']).fillna(0)
                plot_info_non_zero = plot_info[plot_info['feat_imp'] != 0]
                corr = plot_info_non_zero['beta'].astype(float).corr(plot_info_non_zero['feat_imp'].astype(float),
                                                                     method="pearson")
                plot_info_non_zero['beta'] = plot_info_non_zero['beta'].astype(float).abs()
                plot_info_non_zero['feat_imp'] = plot_info_non_zero['feat_imp'].astype(float).abs()
                plot_info_non_zero['feat_imp'] = \
                    (plot_info_non_zero['feat_imp'] - plot_info_non_zero['feat_imp'].min()) / \
                    (plot_info_non_zero['feat_imp'].max() - plot_info_non_zero['feat_imp'].min())
                plot_info_background = plot_info_non_zero[plot_info_non_zero['type'] == 'background']
                plot_info_causal = plot_info_non_zero[plot_info_non_zero['type'] == 'causal']
                alpha_offset = 0.2 if len(models) == 1 else 0
                ax.scatter(x=plot_info_background['feat_imp'], y=plot_info_background['beta'],
                           color=plt.cm.get_cmap("tab10")(color_index/10),
                           alpha=0.3 + alpha_offset, label=model + '(' + str(round(corr, 2)) + ')')
                ax.scatter(x=plot_info_causal['feat_imp'], y=plot_info_causal['beta'],
                           color=plt.cm.get_cmap("tab10")(color_index/10),
                           alpha=0.7, label='_nolegend_', edgecolors='black')
            ax.set(xscale="log", yscale="log")
            ax.set_xlim(0.00001, 1 + np.exp(0.2))
            ax.set_ylim(0.00001, beta_max + 2)
            ax.set_xlabel('Feature importance', fontsize=10)
            ax.set_ylabel('Effect size', fontsize=10)
            ax.set_title(str(sim_id), fontsize=12)
            ax.legend(loc='lower right', fontsize=10, labelspacing=0.2, handletextpad=0.5, borderaxespad=0.2)
            ax.tick_params(labelsize=10)
        plt.subplots_adjust(hspace=0.1)
        suptitle = 'Feature importances of all models versus effect sizes' if len(models) > 1 \
            else 'Feature importances of ' + model + ' versus effect sizes'
        fig.suptitle(suptitle, fontsize=13, horizontalalignment='center')
        fig.tight_layout()
        model_string = models[0] if len(models) == 1 else 'all'
        plt.savefig(
            save_dir.joinpath('Scatterplot_simulated_effect_sizes_vs_featimps_statistics_' + model_string + '_' +
                              datasplit_maf_pattern + '.pdf'), bbox_inches='tight', dpi=300)


def featimps_vs_simulation(results_directory_genotype_level: str, sim_config_dir: str, save_dir: str):
    """
    Analyze feature importances versus effect sizes on synthetic data, both by retrieving stastistics and generating plots

    :param results_directory_genotype_level: results directory at the level of the name of the genotype matrix
    :param sim_config_dir: directory which contains the sim config files
    :param save_dir: directory to save the results
    """

    results_directory_genotype_level = pathlib.Path(results_directory_genotype_level)
    sim_config_dir = pathlib.Path(sim_config_dir)
    save_dir = pathlib.Path(save_dir) if save_dir is not None else results_directory_genotype_level
    sim_configs_gathered_path = save_dir.joinpath('Sim_configs_gathered_' + sim_config_dir.parts[-2] + '.csv')
    if not sim_configs_gathered_path.is_file():
        print('Prepare sim config info')
        gather_sim_configs(sim_config_dir=sim_config_dir, save_dir=save_dir)

    datasplit_maf_patterns = []
    for phenotype_matrix in helper_functions.get_all_subdirectories_non_recursive(results_directory_genotype_level):
        if 'Simulation' not in phenotype_matrix.name:
            continue
        for phenotype_folder in \
                helper_functions.get_all_subdirectories_non_recursive(phenotype_matrix):
            subdirs = [fullpath.parts[-1]
                       for fullpath in helper_functions.get_all_subdirectories_non_recursive(phenotype_folder)]
            datasplit_maf_patterns.extend(['_'.join(path.split('_')[:3]) for path in subdirs])
    datasplit_maf_patterns = set(datasplit_maf_patterns)
    for datasplit_maf_pattern in datasplit_maf_patterns:
        print('Simulation vs Feature Importances for ' + datasplit_maf_pattern)
        feat_imps_gathered_path = save_dir.joinpath('Feature_importances_gathered_' +
                                                    results_directory_genotype_level.parts[-1] +
                                                    datasplit_maf_pattern + '.csv')
        if not feat_imps_gathered_path.is_file():
            print('Prepare feature importance info')
            gather_feature_importances(results_dir=results_directory_genotype_level, save_dir=save_dir,
                                       datasplit_maf_pattern=datasplit_maf_pattern)
        all_sim_configs = pd.read_csv(sim_configs_gathered_path)
        all_feat_imps = pd.read_csv(feat_imps_gathered_path)
        print('Generate and save statistics')
        stats = get_statistics_featimps_vs_simulation(all_sim_configs=all_sim_configs, all_feat_imps=all_feat_imps)
        stats.to_csv(save_dir.joinpath('Simulated_effect_sizes_vs_featimps_statistics_' +
                                       datasplit_maf_pattern + '.csv'), index=False)
        print('Generate and save plots')
        generate_scatterplots_featimps_vs_simulation(all_sim_configs=all_sim_configs, all_feat_imps=all_feat_imps,
                                                     save_dir=save_dir, datasplit_maf_pattern=datasplit_maf_pattern)



import argparse

from . import model_reuse
from ..preprocess import encoding_functions

if __name__ == "__main__":
    """
    Run to train a model on a new dataset using the hyperparameters that worked best for the specified model results.
    """
    parser = argparse.ArgumentParser()
    parser.add_argument("-dd", "--data_dir", type=str,
                        help="Provide the full path of the data directory that contains the geno- and phenotype "
                             "files you want to optimize on")
    parser.add_argument("-gm", "--genotype_matrix", type=str,
                        help="Provide the name of the genotype matrix you want to predict on")
    parser.add_argument("-pm", "--phenotype_matrix", type=str,
                        help="Provide the name of the phenotype matrix you want to predict on")
    parser.add_argument("-p", "--phenotype", type=str,
                        help="Provide the name of the phenotype you want to predict on")
    parser.add_argument("-sd", "--save_dir", type=str,
                        help="Define the save directory for the results.")
    parser.add_argument("-rd", "--results_directory_model", type=str,
                        help="Provide the full path of the directory where your results of the model "
                             "you want to use are stored")

    parser.add_argument("-enc", "--encoding", type=str, default=None,
                        help="specify the encoding to use. Caution: has to be a possible encoding for the model to use."
                             "Valid arguments are: " + str(encoding_functions.get_list_of_encodings()))
    # Preprocess Params #
    parser.add_argument("-maf", "--maf_percentage", type=int, default=0,
                        help="specify the minor allele frequency (as percentage value). "
                             "specify 0 if you do not want a maf filter.")
    parser.add_argument("-split", "--datasplit", type=str, default='nested-cv',
                        help="specify the data split to use: 'nested-cv' | 'cv-test' | 'train-val-test'"
                             "Default values are 5 folds, train-test-split to 80/20 and train-val-test to 60/20/20")
    parser.add_argument("-testperc", "--test_set_size_percentage", type=int, default=20,
                        help="specify the size of the test set in percentage. "
                             "Standard is 20, only relevant for 'cv-test' and 'train-val-test'")
    parser.add_argument("-valperc", "--val_set_size_percentage", type=int, default=20,
                        help="specify the size of the validation set in percentage. "
                             "Standard is 20, only relevant for 'train-val-test'")
    parser.add_argument("-of", "--n_outerfolds", type=int, default=3,
                        help="specify the number of outerfolds to use for 'nested_cv'"
                             "Standard is 5, only relevant for 'nested_cv'")
    parser.add_argument("-folds", "--n_innerfolds", type=int, default=5,
                        help="specify the number of innerfolds/folds to use for 'nested_cv' respectively 'cv-test'"
                             "Standard is 5, only relevant for 'nested_cv' and 'cv-test'")

    # Model and Optimization Params #
    parser.add_argument("-sf", "--save_final_model", default=True,
                        type=lambda x: (str(x).lower() in ['true', '1', 'yes']),
                        help="save the final model to hard drive "
                             "(caution: some models may use a lot of disk space, "
                             "unfitted models that can be retrained are already saved by default)")

    args = vars(parser.parse_args())

    model_reuse.retrain_on_new_data(**args)



import pyro
import pyro.distributions as dist
import torch

from . import _bayesian_linreg


class BayesRidge(_bayesian_linreg.Bayes):
    """
    Implementation of a class for Bayesian Ridge Regression. R implementaiton of Bayes A, B and C is availabile via Docker workflow and weigh faster.

    *Attributes*

        *Inherited attributes*

        See :obj:`~easypheno.model._bayesian_linreg.Bayes` for more information on the attributes.
    """

    def probability_model(self, X: torch.tensor, y: torch.Tensor):
        """
        Implementation of probability model for Bayesian ridge regression

        See :obj:`~easypheno.model._bayesian_linreg.Bayes` for more information on the attributes.
        """
        # Define our intercept prior
        intercept_prior = dist.Normal(0.0, 1.0)
        linear_combination = pyro.sample(f"beta_intercept", intercept_prior)

        # Also define coefficient priors
        for i in range(X.shape[1]):
            coefficient_prior = dist.Normal(0.0, 1.0)
            beta_coef = pyro.sample(f"beta_{i}", coefficient_prior)
            linear_combination = linear_combination + (X[:, i] * beta_coef)

        # Define a sigma prior for the random error
        sigma = pyro.sample("sigma", dist.HalfNormal(scale=10.0))

        # For a simple linear model, the expected mean is the linear combination of parameters
        mean = linear_combination

        with pyro.plate("data", y.shape[0]):
            # Assume our expected mean comes from a normal distribution with the mean which
            # depends on the linear combination, and a standard deviation "sigma"
            outcome_dist = dist.Normal(mean, torch.clamp(sigma, min=0.0001, max=10**6))

            # Condition the expected mean on the observed target y
            observation = pyro.sample("obs", outcome_dist, obs=y)



import abc
import joblib
import numpy as np
import pathlib


class ParamFreeBaseModel(abc.ABC):
    """
    BaseModel parent class for all models that do not have hyperparameters, e.g. BLUP.

    Every model must be based on :obj:`~easypheno.model.param_free_base_model.ParamFreeBaseModel` directly or ParamFreeBaseModel's child classes.

    Please add ``super().__init__(PARAMS)`` to the constructor in case you override it in a child class

    **Attributes**

        *Class attributes*

        - standard_encoding (*str*): the standard encoding for this model
        - possible_encodings (*List<str>*): a list of all encodings that are possible according to the model definition

        *Instance attributes*

        - task (*str*): ML task ('regression' or 'classification') depending on target variable
        - encoding (*str*): the encoding to use (standard encoding or user-defined)


    :param task: ML task (regression or classification) depending on target variable
    :param encoding: the encoding to use (standard encoding or user-defined)

    """

    # Class attributes #
    @property
    @classmethod
    @abc.abstractmethod
    def standard_encoding(cls):
        """the standard encoding for this model"""
        raise NotImplementedError

    @property
    @classmethod
    @abc.abstractmethod
    def possible_encodings(cls):
        """a list of all encodings that are possible according to the model definition"""
        raise NotImplementedError

    # Constructor super class #
    def __init__(self, task: str, encoding: str = None):
        self.task = task
        self.encoding = self.standard_encoding if encoding is None else encoding

    # Methods required by each child class #

    @abc.abstractmethod
    def fit(self, X: np.array, y: np.array) -> np.array:
        """
        Method that fits the model based on features X and targets y

        :param X: feature matrix for retraining
        :param y: target vector

        :return: numpy array with values predicted for X
        """

    @abc.abstractmethod
    def predict(self, X_in: np.array) -> np.array:
        """
        Method that predicts target values based on the input X_in

        :param X_in: feature matrix as input

        :return: numpy array with the predicted values
        """

    def save_model(self, path: pathlib.Path, filename: str):
        """
        Persist the whole model object on a hard drive
        (can be loaded with :obj:`~easypheno.model._model_functions.load_model`)

        :param path: path where the model will be saved
        :param filename: filename of the model
        """
        joblib.dump(self, path.joinpath(filename), compress=3)



import xgboost

from . import _sklearn_model


class XgBoost(_sklearn_model.SklearnModel):
    """
    Implementation of a class for XGBoost.

    See :obj:`~easypheno.model._base_model.BaseModel` for more information on the attributes.
    """
    standard_encoding = '012'
    possible_encodings = ['012']

    def define_model(self) -> xgboost.XGBModel:
        """
        Definition of the actual prediction model.

        See :obj:`~easypheno.model._base_model.BaseModel` for more information.
        """
        # all hyperparameters defined for XGBoost are suggested for optimization
        params = self.suggest_all_hyperparams_to_optuna()
        # add random_state for reproducibility
        params.update({'random_state': 42, 'reg_lambda': 0})
        if self.task == 'classification':
            # set some parameters to prevent warnings
            params.update({'use_label_encoder': False})
            eval_metric = 'mlogloss' if self.n_outputs > 2 else 'logloss'
            params.update({'eval_metric': eval_metric})
            return xgboost.XGBClassifier(**params)
        else:
            return xgboost.XGBRegressor(**params)

    def define_hyperparams_to_tune(self) -> dict:
        """
        See :obj:`~easypheno.model._base_model.BaseModel` for more information on the format.

        Further params that potentially can be optimized

            .. code-block:: python

                'reg_lambda': {
                    'datatype': 'float',
                    'lower_bound': 0,
                    'upper_bound': 1000,
                    'step': 10
                }

        """
        return {
            'n_estimators': {
                'datatype': 'categorical',
                'list_of_values': [50, 100, 250, 500, 750, 1000, 1250, 1500, 1750, 2000, 2250, 2500, 2750, 3000]
            },
            'learning_rate': {
                    'datatype': 'float',
                    'lower_bound': 0.025,
                    'upper_bound': 0.3,
                    'step': 0.025
            },
            'max_depth': {
                'datatype': 'int',
                'lower_bound': 2,
                'upper_bound': 10
            },
            'gamma': {
                'datatype': 'int',
                'lower_bound': 0,
                'upper_bound': 1000,
                'step': 10
            },
            'subsample': {
                'datatype': 'float',
                'lower_bound': 0.05,
                'upper_bound': 0.8,
                'step': 0.05
            },
            'colsample_bytree': {
                'datatype': 'float',
                'lower_bound': 0.05,
                'upper_bound': 0.8,
                'step': 0.05
            },
            'reg_alpha': {
                'datatype': 'float',
                'lower_bound': 0,
                'upper_bound': 1000,
                'step': 10
            }
        }



import datetime
import pprint
import pathlib
import warnings
from optuna.exceptions import ExperimentalWarning

from easypheno.utils import check_functions, print_functions, helper_functions
from easypheno.preprocess import raw_data_functions, base_dataset
from easypheno.optimization import optuna_optim, paramfree_fitting
from easypheno.model import _param_free_base_model
import easypheno.model


def run(data_dir: str, genotype_matrix: str, phenotype_matrix: str, phenotype: str,
        encoding: str = None, maf_percentage: int = 0, save_dir: str = None,
        datasplit: str = 'nested-cv', n_outerfolds: int = 5, n_innerfolds: int = 5,
        test_set_size_percentage: int = 20, val_set_size_percentage: int = 20,
        models: list = None, n_trials: int = 100, save_final_model: bool = False,
        batch_size: int = 32, n_epochs: int = 100000, outerfold_number_to_run: int = None):
    """
    Run the whole optimization pipeline

    :param data_dir: data directory where the phenotype and genotype matrix are stored
    :param genotype_matrix: name of the genotype matrix including datatype ending
    :param phenotype_matrix: name of the phenotype matrix including datatype ending
    :param phenotype: name of the phenotype to predict
    :param encoding: encoding to use. Default is None, so standard encoding of each model will be used. Options are: '012', 'onehot', 'raw'
    :param maf_percentage: threshold for MAF filter as percentage value. Default is 0, so no MAF filtering
    :param save_dir: directory for saving the results. Default is None, so same directory as data_dir
    :param datasplit: datasplit to use. Options are: nested-cv, cv-test, train-val-test
    :param n_outerfolds: number of outerfolds relevant for nested-cv
    :param n_innerfolds: number of folds relevant for nested-cv and cv-test
    :param test_set_size_percentage: size of the test set relevant for cv-test and train-val-test
    :param val_set_size_percentage: size of the validation set relevant for train-val-test
    :param models: list of models that should be optimized
    :param n_trials: number of trials for optuna
    :param save_final_model: specify if the final model should be saved
    :param batch_size: batch size for neural network models
    :param n_epochs: number of epochs for neural network models
    :param outerfold_number_to_run: outerfold to run in case you do not want to run all
    """
    warnings.simplefilter(action='ignore', category=FutureWarning)
    warnings.simplefilter(action='ignore', category=ExperimentalWarning)

    if models is None:
        models = ['xgboost']
    if any(['fromR' in model for model in models]):
        easypheno.model.__all__.extend(['_bayesfromR', 'bayesAfromR', 'bayesBfromR', 'bayesCfromR'])
    # create Path
    data_dir = pathlib.Path(data_dir)
    # set save directory
    save_dir = data_dir if save_dir is None else pathlib.Path(save_dir)
    save_dir = save_dir if save_dir.is_absolute() else save_dir.resolve()
    if type(models) == list and models[0] == 'all':
        models = 'all'
    if type(models) != list and models != 'all':
        models = [models]

    # Checks and Raw Data Input Preparation #
    # Check all arguments
    check_functions.check_all_specified_arguments(arguments=locals())
    # prepare all data files
    raw_data_functions.prepare_data_files(
        data_dir=data_dir, genotype_matrix_name=genotype_matrix, phenotype_matrix_name=phenotype_matrix,
        phenotype=phenotype, datasplit=datasplit, n_outerfolds=n_outerfolds, n_innerfolds=n_innerfolds,
        test_set_size_percentage=test_set_size_percentage, val_set_size_percentage=val_set_size_percentage,
        models=models, user_encoding=encoding, maf_percentage=maf_percentage
    )

    # Optimization Pipeline #
    helper_functions.set_all_seeds()
    models_to_optimize = helper_functions.get_list_of_implemented_models() if models == 'all' else models
    if len(models_to_optimize) > 1:
        models_to_optimize = helper_functions.sort_models_by_encoding(models_list=models_to_optimize)
    model_overview = {}
    only_ofn_postfix = '' if outerfold_number_to_run is None else 'Outerfold' + str(outerfold_number_to_run)
    models_start_time = \
        '+'.join(models_to_optimize) + '_' + datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S") + only_ofn_postfix
    user_encoding = encoding
    for optim_run, current_model_name in enumerate(models_to_optimize):
        encoding = user_encoding if user_encoding is not None \
            else helper_functions.get_mapping_name_to_class()[current_model_name].standard_encoding
        if optim_run == 0:
            print('----- Starting dataset preparation -----')
            dataset = base_dataset.Dataset(
                data_dir=data_dir, genotype_matrix_name=genotype_matrix, phenotype_matrix_name=phenotype_matrix,
                phenotype=phenotype, datasplit=datasplit, n_outerfolds=n_outerfolds, n_innerfolds=n_innerfolds,
                test_set_size_percentage=test_set_size_percentage, val_set_size_percentage=val_set_size_percentage,
                encoding=encoding, maf_percentage=maf_percentage
            )
            task = 'classification' if helper_functions.test_likely_categorical(dataset.y_full) else 'regression'
            print_functions.print_config_info(arguments=locals(), dataset=dataset, task=task)
        else:
            if dataset.encoding != encoding:
                print('----- Load new dataset encoding -----')
                dataset = base_dataset.Dataset(
                    data_dir=data_dir, genotype_matrix_name=genotype_matrix, phenotype_matrix_name=phenotype_matrix,
                    phenotype=phenotype, datasplit=datasplit, n_outerfolds=n_outerfolds, n_innerfolds=n_innerfolds,
                    test_set_size_percentage=test_set_size_percentage, val_set_size_percentage=val_set_size_percentage,
                    encoding=encoding, maf_percentage=maf_percentage
                )
        if issubclass(helper_functions.get_mapping_name_to_class()[current_model_name],
                      _param_free_base_model.ParamFreeBaseModel):
            print('### Starting Model Fitting for ' + current_model_name + ' ###')
            optim_run = paramfree_fitting.ParamFreeFitting(
                save_dir=save_dir, genotype_matrix_name=genotype_matrix, phenotype_matrix_name=phenotype_matrix,
                phenotype=phenotype, n_outerfolds=n_outerfolds, n_innerfolds=n_innerfolds,
                val_set_size_percentage=val_set_size_percentage, test_set_size_percentage=test_set_size_percentage,
                maf_percentage=maf_percentage, save_final_model=save_final_model,
                task=task, models_start_time=models_start_time, current_model_name=current_model_name, dataset=dataset
            )
            overall_results = optim_run.run_fitting()
            print('### Finished Model Fitting for ' + current_model_name + ' ###')
        else:
            optim_run = optuna_optim.OptunaOptim(
                save_dir=save_dir, genotype_matrix_name=genotype_matrix, phenotype_matrix_name=phenotype_matrix,
                phenotype=phenotype, n_outerfolds=n_outerfolds, n_innerfolds=n_innerfolds,
                val_set_size_percentage=val_set_size_percentage, test_set_size_percentage=test_set_size_percentage,
                maf_percentage=maf_percentage, n_trials=n_trials, save_final_model=save_final_model,
                batch_size=batch_size, n_epochs=n_epochs, task=task, models_start_time=models_start_time,
                current_model_name=current_model_name, dataset=dataset, outerfold_number_to_run=outerfold_number_to_run)
            print('### Starting Optuna Optimization for ' + current_model_name + ' ###')
            overall_results = optim_run.run_optuna_optimization()
            print('### Finished Optuna Optimization for ' + current_model_name + ' ###')
        model_overview[current_model_name] = overall_results
    print('# Optimization runs done for models ' + str(models_to_optimize))
    print('Results overview on the test set(s)')
    pprint.PrettyPrinter(depth=4).pprint(model_overview)
    path_overview_file = optim_run.base_path.parent.joinpath(
        'Results_overview_' + '_'.join(models) + only_ofn_postfix + '.csv')
    helper_functions.save_model_overview_dict(model_overview=model_overview, save_path=path_overview_file)



import numpy as np
import torch
from torch.nn.functional import one_hot

from ..utils import helper_functions


def get_encoding(models, user_encoding: str) -> list:
    """
    Get a list of all required encodings.

    :param models: models to consider
    :param user_encoding: encoding specified by the user

    :return: list of encodings
    """
    if models == 'all':
        # do all encodings if all models are specified
        list_of_encodings = get_list_of_encodings()
    else:
        # use the specified encoding or encodings of specified models
        if user_encoding is not None:
            list_of_encodings = [user_encoding]
        else:
            list_of_encodings = []
            for model in models:
                if helper_functions.get_mapping_name_to_class()[model].standard_encoding not in list_of_encodings:
                    list_of_encodings.append(helper_functions.get_mapping_name_to_class()[model].standard_encoding)
    return list_of_encodings


def get_list_of_encodings() -> list:
    """
    Get a list of all implemented encodings.

    ! Adapt if new encoding is added !

    :return: List of all possible encodings
    """
    return ['raw', '012', 'onehot', '101']


def get_base_encoding(encoding: str) -> str:
    """
    Check which base encoding is needed to create required encoding.

    ! Adapt if new encoding is added !

    :param encoding: required encoding

    :return: base encoding
    """
    if encoding in ('raw', '012', 'onehot', '101'):
        return 'raw'
    else:
        raise Exception('No valid encoding. Can not determine base encoding')


def check_encoding_of_genotype(X: np.array) -> str:
    """
    Check the encoding of the genotype matrix

    ! Adapt if new encoding is added !

    :param X: genotype matrix

    :return: encoding of the genotype matrix
    """
    unique = np.unique(X)
    if all(z in ['A', 'C', 'G', 'T', 'M', 'R', 'W', 'S', 'Y', 'K'] for z in unique.astype(str)):
        return 'raw'
    elif all(z in [0, 1, 2] for z in unique):
        return '012'
    elif all(z in ["AA", "GG", "TT", "CC", "AG", "GA", "CT", "TC", "GC",
                   "CG", "AT", "TA", "GT", "TG", "AC", "CA"] for z in unique.astype(str)):
        return 'biallelic'


def encode_genotype(X: np.array, required_encoding: str) -> np.array:
    """
    Compute the required encoding of the genotype matrix

    ! Adapt if new encoding is added !

    :param X: genotype matrix
    :param required_encoding: encoding of genotype matrix to create

    :return: X in new encoding
    """
    if required_encoding == '012':
        return get_additive_encoding(X)
    elif required_encoding == 'onehot':
        return get_onehot_encoding(X)
    elif required_encoding == '101':
        return get_additive_encoding(X, '101')
    else:
        raise Exception('Only able to create additive or onehot encoding.')


def get_additive_encoding(X: np.array, style: str = '012') -> np.array:
    """
    Generate genotype matrix in additive encoding:

    - 0: homozygous major allele,
    - 1: heterozygous
    - 2: homozygous minor allele
    for style=012
    - 1: homozygous major allele,
    - 0: heterozygous
    - -1: homozygous minor allele

    :param X: genotype matrix in raw encoding, i.e. containing the alleles
    :param style: encoding style, '012' or '101' default is '012'

    :return: genotype matrix in additive encoding (X_012)
    """
    if style == '012':
        shift = 0
    elif style == '101':
        shift = 1
    else:
        raise Exception('only allow 012 or 101 as style for additive encoding')
    alleles = []
    index_arr = []
    pairs = [['A', 'C'], ['A', 'G'], ['A', 'T'], ['C', 'G'], ['C', 'T'], ['G', 'T']]
    heterozygous_nuc = ['M', 'R', 'W', 'S', 'Y', 'K']
    for j, col in enumerate(np.transpose(X)):
        unique, inv, counts = np.unique(col, return_counts=True, return_inverse=True)
        unique = unique.astype(str)
        boolean = (unique == 'A') | (unique == 'T') | (unique == 'C') | (unique == 'G')
        tmp = np.zeros(3) + shift
        if len(unique) > 3:
            raise Exception('More than two alleles encountered at snp ' + str(j))
        elif len(unique) == 3:
            hetero = unique[~boolean][0]
            homozygous = unique[boolean]
            for i, pair in enumerate(pairs):
                if all(h in pair for h in homozygous) and hetero != heterozygous_nuc[i]:
                    raise Exception('More than two alleles encountered at snp ' + str(i))
            tmp[~boolean] = 1.0 - shift
            tmp[np.argmin(counts[boolean])] = 2.0 - 3*shift
        elif len(unique) == 2:
            if list(unique) in pairs:
                tmp[np.argmin(counts)] = 2.0 - 3*shift
            else:
                tmp[(~boolean).nonzero()] = 1.0 - shift
        else:
            if unique[0] in heterozygous_nuc:
                tmp[0] = 1.0 - shift
        alleles.append(tmp)
        index_arr.append(inv)
    alleles = np.transpose(np.array(alleles))
    ind_arr = np.transpose(np.array(index_arr))
    cols = np.arange(alleles.shape[1])
    return alleles[ind_arr, cols]


def get_onehot_encoding(X: np.array) -> np.array:
    """
    Generate genotype matrix in onehot encoding. If genotype matrix is homozygous, create 3d torch tensor with
    (samples, SNPs, 4), with 4 as the onehot encoding

    - A : [1,0,0,0]
    - C : [0,1,0,0]
    - G : [0,0,1,0]
    - T : [0,0,0,1]

    If genotype matrix is heterozygous, create 3d torch tensor with (samples, SNPs, 10), with 10 as the onehot encoding

    - A : [1,0,0,0,0,0,0,0,0,0]
    - C : [0,1,0,0,0,0,0,0,0,0]
    - G : [0,0,1,0,0,0,0,0,0,0]
    - K : [0,0,0,1,0,0,0,0,0,0]
    - M : [0,0,0,0,1,0,0,0,0,0]
    - R : [0,0,0,0,0,1,0,0,0,0]
    - S : [0,0,0,0,0,0,1,0,0,0]
    - T : [0,0,0,0,0,0,0,1,0,0]
    - W : [0,0,0,0,0,0,0,0,1,0]
    - Y : [0,0,0,0,0,0,0,0,0,1]

    :param X: genotype matrix in raw encoding, i.e. containing the alleles

    :return: genotype matrix in onehot encoding (X_onehot)
    """
    unique, inverse = np.unique(X, return_inverse=True)
    inverse = inverse.reshape(X.shape)
    X_onehot = one_hot(torch.from_numpy(inverse)).numpy()
    return X_onehot



import numpy as np
import pandas as pd
import pyro
import torch

from . import _param_free_base_model


class Bayes(_param_free_base_model.ParamFreeBaseModel):
    """
    Implementation of a class for Bayesian linear regression.

    *Attributes*

        *Inherited attributes*

        See :obj:`~easypheno.model._param_free_base_model.ParamFreeBaseModel` for more information on the attributes.

        *Additional attributes*

        - mu (*np.array*): intercept
        - beta (*np.array*): effect size
        - iterations (*int*): MCMC sampler iterations
        - warmup (*int*): number of discarded MCMC warmup iterations
    """
    standard_encoding = '012'
    possible_encodings = ['101']

    def __init__(self, task: str, encoding: str = None, iterations: int = 100, warmup: int = 10):
        super().__init__(task=task, encoding=encoding)
        self.mu = None
        self.beta = None
        self.iterations = iterations
        self.warmup = warmup

    def probability_model(self, X: torch.Tensor, y: torch.Tensor):
        """
        Probability model that needs to be implemented by each child model.

        :param X: feature matrix
        :param y: target vector
        """
        raise NotImplementedError

    def fit(self, X: np.array, y: np.array) -> np.array:
        """
        Implementation of fit function for Bayesian linear regression.

        See :obj:`~easypheno.model._param_free_base_model.ParamFreeBaseModel` for more information.
        """
        torch.autograd.set_detect_anomaly(True)
        X = torch.tensor(X)
        y = torch.tensor(y)

        # Clear the parameter storage
        pyro.clear_param_store()

        # Initialize our No U-Turn Sampler
        my_kernel = pyro.infer.NUTS(self.probability_model, max_tree_depth=7)  # a shallower tree helps the algorithm run faster

        # Employ the sampler in an MCMC sampling
        # algorithm, and sample 3100 samples.
        # Then discard the first 100
        mcmc = pyro.infer.MCMC(my_kernel, num_samples=self.iterations, warmup_steps=self.warmup)
        # Run the sampler
        mcmc.run(X, y)

        # save results as numpy arrays
        coefficients = pd.DataFrame(mcmc.get_samples()).iloc[:, :-1].mean()
        self.beta = coefficients[1:]
        self.mu = coefficients.iloc[0]

        return self.predict(X_in=X)

    def predict(self, X_in: np.array) -> np.array:
        """
        Implementation of predict function for Bayesian linear regression.

        See :obj:`~easypheno.model._param_free_base_model.ParamFreeBaseModel` for more information.
        """
        return self.mu + np.matmul(X_in, self.beta)



import numpy as np
import joblib
import tensorflow as tf
import pathlib
import pandas as pd
import optuna

from . import _base_model, _tensorflow_model, _param_free_base_model, _torch_model
from ..evaluation import eval_metrics
from ..utils import helper_functions
from ..postprocess import results_analysis
from ..preprocess import base_dataset


def load_retrain_model(path: pathlib.Path, filename: str, X_retrain: np.array, y_retrain: np.array,
                       early_stopping_point: int = None) -> _base_model.BaseModel:
    """
    Load and retrain persisted model

    :param path: path where the model is saved
    :param filename: filename of the model
    :param X_retrain: feature matrix for retraining
    :param y_retrain: target vector for retraining
    :param early_stopping_point: optional early stopping point relevant for some models

    :return: model instance
    """
    model = load_model(path=path, filename=filename)
    if early_stopping_point is not None:
        model.early_stopping_point = early_stopping_point
    model.retrain(X_retrain=X_retrain, y_retrain=y_retrain)
    return model


def retrain_model_with_results_file(results_file_path: pathlib.Path, model_name: str, datasplit: str,
                                    outerfold_number: int, dataset: base_dataset.Dataset,
                                    saved_outerfold_number: int = None, saved_datasplit: str = None) \
        -> _base_model.BaseModel:
    """
    Retrain a model based on information saved in a results overview file.

    :param results_file_path: path to the results overview file containing the information for retraining
    :param model_name: name of the model to retrain
    :param datasplit: datasplit to use
    :param outerfold_number: outerfold number to use
    :param dataset: dataset for training
    :param saved_outerfold_number: outerfold number of the results file in case the model is retrained with fixed hyperparameters of one outerfold
    :param saved_datasplit: outerfold number of the results file in case the model is retrained on another datasplit

    :return: retrained model
    """

    saved_outerfold_number = outerfold_number if saved_outerfold_number is None else saved_outerfold_number
    saved_datasplit = datasplit if saved_datasplit is None else saved_datasplit
    outerfold_info = dataset.datasplit_indices['outerfold_' + str(outerfold_number)]
    X_test, y_test, sample_ids_test = \
        dataset.X_full[outerfold_info['test']], dataset.y_full[outerfold_info['test']], \
        dataset.sample_ids_full[outerfold_info['test']]
    X_retrain, y_retrain, sample_ids_retrain = \
        dataset.X_full[~np.isin(np.arange(len(dataset.X_full)), outerfold_info['test'])], \
        dataset.y_full[~np.isin(np.arange(len(dataset.y_full)), outerfold_info['test'])], \
        dataset.sample_ids_full[~np.isin(np.arange(len(dataset.sample_ids_full)), outerfold_info['test'])]

    results = pd.read_csv(results_file_path)
    results = results[results[results.columns[0]] == 'outerfold_' + str(saved_outerfold_number)] \
        if saved_datasplit == 'nested-cv' else results
    results = results.loc[:, [model_name in col for col in results.columns]]
    eval_dict_saved = results_analysis.result_string_to_dictionary(
        result_string=results[model_name + '___eval_metrics'][saved_outerfold_number]
    )
    task = 'regression' if 'test_rmse' in eval_dict_saved.keys() else 'classification'
    helper_functions.set_all_seeds()
    if issubclass(helper_functions.get_mapping_name_to_class()[model_name],
                  _param_free_base_model.ParamFreeBaseModel):
        model: _param_free_base_model.ParamFreeBaseModel = \
            helper_functions.get_mapping_name_to_class()[model_name](
                task=task,
            )
        _ = model.fit(X=X_retrain, y=y_retrain)
    else:
        best_params = results_analysis.result_string_to_dictionary(
            result_string=results[model_name + '___best_params'][saved_outerfold_number]
        )
        trial = optuna.trial.FixedTrial(params=best_params)
        additional_attributes_dict = {}
        if issubclass(helper_functions.get_mapping_name_to_class()[model_name], _torch_model.TorchModel) or \
                issubclass(helper_functions.get_mapping_name_to_class()[model_name], _tensorflow_model.TensorflowModel):
            # additional attributes for torch and tensorflow models
            additional_attributes_dict['n_features'] = dataset.X_full.shape[1]
            additional_attributes_dict['width_onehot'] = dataset.X_full.shape[-1]
            additional_attributes_dict['early_stopping_point'] = best_params['early_stopping_point']
            additional_attributes_dict['n_epochs'] = best_params['n_epochs']
            additional_attributes_dict['batch_size'] = best_params['batch_size']
        model: _base_model.BaseModel = helper_functions.get_mapping_name_to_class()[model_name](
            task=task, optuna_trial=trial,
            n_outputs=len(np.unique(dataset.y_full)) if task == 'classification' else 1,
            **additional_attributes_dict
        )
        model.retrain(X_retrain=X_retrain, y_retrain=y_retrain)
    y_pred_test = model.predict(X_in=X_test)
    eval_scores = \
        eval_metrics.get_evaluation_report(y_pred=y_pred_test, y_true=y_test,
                                           task=model.task,
                                           prefix='test_')
    print('Compare Results from initial testing to refitting')
    print('New fitting: ')
    print(eval_scores)
    print('Old fitting: ')
    print(eval_dict_saved)

    return model


def load_model(path: pathlib.Path, filename: str) -> _base_model.BaseModel:
    """
    Load persisted model

    :param path: path where the model is saved
    :param filename: filename of the model

    :return: model instance
    """
    model = joblib.load(path.joinpath(filename))
    # special case for loading tensorflow optimizer
    if issubclass(type(model), _tensorflow_model.TensorflowModel):
        model.optimizer = tf.keras.optimizers.deserialize(model.optimizer)
    return model




import os

import pandas as pd
import pathlib

from ..preprocess import base_dataset
from ..utils import helper_functions, check_functions
from ..model import _model_functions


def post_generate_feature_importances(results_directory_genotype_level: str, data_dir: str):
    """
    Post-generate the feature importances for several models for all sub-folders of the specified directory of already optimized models.
    Only needed in case you e.g. forgot to implement the saving of the feature importances.

    :param results_directory_genotype_level: results directory at the level of the name of the genotype matrix
    :param data_dir: data directory where the phenotype and genotype matrix as well as index file are stored
    """
    results_directory_genotype_level = pathlib.Path(results_directory_genotype_level)
    data_dir = pathlib.Path(data_dir)

    # Check user inputs
    print("Checking user inputs")
    if not check_functions.check_exist_directories(list_of_dirs=[results_directory_genotype_level, data_dir]):
        raise Exception("See output above. Problems with specified directories")
    if not data_dir.joinpath(results_directory_genotype_level.parts[-1] + '.h5').is_file():
        raise Exception("Genotype matrix specified does not exist: " + str(results_directory_genotype_level) +
                        "\n Make sure the results directory is at the level fo the genotype matrix name.")

    genotype_name = results_directory_genotype_level.parts[-1] + '.h5'
    for phenotype_matrix in helper_functions.get_all_subdirectories_non_recursive(results_directory_genotype_level):
        study_name = phenotype_matrix.parts[-1] + '.csv'
        for phenotype_folder in \
                helper_functions.get_all_subdirectories_non_recursive(phenotype_matrix):
            print('++++++++++++++ PHENOTYPE ' + phenotype_folder.parts[-1] + ' ++++++++++++++')
            subdirs = [fullpath.parts[-1]
                       for fullpath in helper_functions.get_all_subdirectories_non_recursive(phenotype_folder)]
            datasplit_maf_patterns = set(['_'.join(path.split('_')[:3]) for path in subdirs])
            for pattern in list(datasplit_maf_patterns):
                datasplit, n_outerfolds, n_innerfolds, val_set_size_percentage, test_set_size_percentage, maf_perc = \
                    helper_functions.get_datasplit_config_info_for_resultfolder(resultfolder=pattern)
                dataset = base_dataset.Dataset(
                    data_dir=data_dir, genotype_matrix_name=genotype_name, phenotype_matrix_name=study_name,
                    phenotype=phenotype_folder.parts[-1],
                    datasplit=datasplit, n_outerfolds=n_outerfolds, n_innerfolds=n_innerfolds,
                    test_set_size_percentage=test_set_size_percentage, val_set_size_percentage=val_set_size_percentage,
                    encoding='012', maf_percentage=maf_perc
                )
                # CAUTION: currently the '012' encoding works for all algos with featimps, may need reimplementation
                snp_ids_df = pd.DataFrame(dataset.snp_ids)
                print('Saving SNP ids')
                snp_ids_df.to_csv(
                    phenotype_folder.joinpath('snp_ids.csv'),
                    sep=',', decimal='.', float_format='%.10f',
                    index=False
                )

                for outerfold in range(n_outerfolds):
                    print('Working on outerfold ' + str(outerfold))
                    # Retrain on full train + val data with best hyperparams and apply on test
                    print("## Retrain best model and test ##")
                    for path in phenotype_folder.glob(pattern + '*'):
                        models = path.parts[-1].split('_')[3].split('+')
                        print('working on ' + str(path))
                        for current_model in models:
                            print('Model: ' + current_model)
                            if current_model in ['randomforest', 'xgboost', 'linearregression', 'elasticnet', 'blup']:
                                current_directory = path.joinpath(current_model) if datasplit != 'nested-cv' \
                                    else path.joinpath('outerfold_' + str(outerfold), current_model)
                                if os.path.exists(current_directory.joinpath('final_model_feature_importances.csv')):
                                    print('Already existing')
                                    continue
                                results_file_path = path.joinpath('Results_overview_' + '_'.join(models) + '.csv')
                                if not results_file_path.is_file():
                                    print('No results file')
                                    continue
                                model = _model_functions.retrain_model_with_results_file(
                                    results_file_path=results_file_path, model_name=current_model, datasplit=datasplit,
                                    outerfold_number=outerfold, dataset=dataset
                                )
                                top_n = min(len(dataset.snp_ids), 1000)
                                feat_import_df = pd.DataFrame()
                                if current_model in ['randomforest', 'xgboost']:
                                    feature_importances = model.model.feature_importances_
                                    sorted_idx = feature_importances.argsort()[::-1][:top_n]
                                    feat_import_df['snp_ids_standard'] = dataset.snp_ids[sorted_idx]
                                    feat_import_df['feat_importance_standard'] = feature_importances[sorted_idx]
                                elif current_model in ['linearregression', 'elasticnet']:
                                    coefs = model.model.coef_
                                    dims = coefs.shape[0] if len(coefs.shape) > 1 else 1
                                    for dim in range(dims):
                                        coef = coefs[dim] if len(coefs.shape) > 1 else coefs
                                        sorted_idx = coef.argsort()[::-1][:top_n]
                                        feat_import_df['snp_ids_' + str(dim)] = dataset.snp_ids[sorted_idx]
                                        feat_import_df['coefficients_' + str(dim)] = coef[sorted_idx]
                                else:
                                    feat_imps = model.u if current_model == 'blup' else model.beta
                                    dims = 1
                                    for dim in range(dims):
                                        coef = feat_imps.flatten()
                                        sorted_idx = coef.argsort()[::-1][:top_n]
                                        feat_import_df['snp_ids_' + str(dim)] = dataset.snp_ids[sorted_idx]
                                        feat_import_df['coefficients_' + str(dim)] = coef[sorted_idx]
                                feat_import_df.to_csv(
                                    current_directory.joinpath('final_model_feature_importances.csv'),
                                    sep=',', decimal='.', float_format='%.10f',
                                    index=False
                                )



import abc
import optuna
import joblib
import numpy as np
import pathlib


class BaseModel(abc.ABC):
    """
    BaseModel parent class for all models that can be used within the framework.

    Every model must be based on :obj:`~easypheno.model._base_model.BaseModel` directly or BaseModel's child classes,
    e.g. :obj:`~easypheno.model._sklearn_model.SklearnModel` or :obj:`~easypheno.model._torch_model.TorchModel`

    Please add ``super().__init__(PARAMS)`` to the constructor in case you override it in a child class

    **Attributes**

        *Class attributes*

        - standard_encoding (*str*): the standard encoding for this model
        - possible_encodings (*List<str>*): a list of all encodings that are possible according to the model definition

        *Instance attributes*

        - task (*str*): ML task ('regression' or 'classification') depending on target variable
        - optuna_trial (*optuna.trial.Trial*): trial of optuna for optimization
        - encoding (*str*): the encoding to use (standard encoding or user-defined)
        - n_outputs (*int*): number of outputs of the prediction model
        - all_hyperparams (*dict*): dictionary with all hyperparameters with related info that can be tuned (structure see :obj:`~easypheno.model._base_model.BaseModel.define_hyperparams_to_tune`)
        - model: model object


    :param task: ML task (regression or classification) depending on target variable
    :param optuna_trial: Trial of optuna for optimization
    :param encoding: the encoding to use (standard encoding or user-defined)
    :param n_outputs: number of outputs of the prediction model

    """

    # Class attributes #
    @property
    @classmethod
    @abc.abstractmethod
    def standard_encoding(cls):
        """the standard encoding for this model"""
        raise NotImplementedError

    @property
    @classmethod
    @abc.abstractmethod
    def possible_encodings(cls):
        """a list of all encodings that are possible according to the model definition"""
        raise NotImplementedError

    # Constructor super class #
    def __init__(self, task: str, optuna_trial: optuna.trial.Trial, encoding: str = None, n_outputs: int = 1):
        self.task = task
        self.encoding = self.standard_encoding if encoding is None else encoding
        self.optuna_trial = optuna_trial
        self.n_outputs = n_outputs if task == 'classification' else 1
        if not hasattr(self, 'all_hyperparams'):
            self.all_hyperparams = self.define_hyperparams_to_tune()
        else:
            # update in case common hyperparams are already defined
            self.all_hyperparams.update(self.define_hyperparams_to_tune())
        self.model = self.define_model()

    # Methods required by each child class #
    @abc.abstractmethod
    def define_model(self):
        """
        Method that defines the model that needs to be optimized.
        Hyperparams to tune have to be specified in all_hyperparams and suggested via suggest_hyperparam_to_optuna().
        The hyperparameters have to be included directly in the model definiton to be optimized.
        e.g. if you want to optimize the number of layers, do something like

            .. code-block:: python

                n_layers = self.suggest_hyperparam_to_optuna('n_layers') # same name in define_hyperparams_to_tune()
                for layer in n_layers:
                    do something

        Then the number of layers will be optimized by optuna.
        """

    @abc.abstractmethod
    def define_hyperparams_to_tune(self) -> dict:
        """
        Method that defines the hyperparameters that should be tuned during optimization and their ranges.
        Required format is a dictionary with:

        .. code-block:: python

            {
                'name_hyperparam_1':
                    {
                    # MANDATORY ITEMS
                    'datatype': 'float' | 'int' | 'categorical',
                    FOR DATATYPE 'categorical':
                        'list_of_values': []  # List of all possible values
                    FOR DATATYPE ['float', 'int']:
                        'lower_bound': value_lower_bound,
                        'upper_bound': value_upper_bound,
                        # OPTIONAL ITEMS (only for ['float', 'int']):
                        'log': True | False  # sample value from log domain or not
                        'step': step_size # step of discretization.
                                            # Caution: cannot be combined with log=True
                                                            # - in case of 'float' in general and
                                                            # - for step!=1 in case of 'int'
                    },
                'name_hyperparam_2':
                    {
                    ...
                    },
                ...
                'name_hyperparam_k':
                    {
                    ...
                    }
            }

        If you want to use a similar hyperparameter multiple times (e.g. Dropout after several layers),
        you only need to specify the hyperparameter once. Individual parameters for every suggestion will be created.
        """

    @abc.abstractmethod
    def retrain(self, X_retrain: np.array, y_retrain: np.array):
        """
        Method that runs the retraining of the model

        :param X_retrain: feature matrix for retraining
        :param y_retrain: target vector for retraining
        """

    @abc.abstractmethod
    def predict(self, X_in: np.array) -> np.array:
        """
        Method that predicts target values based on the input X_in

        :param X_in: feature matrix as input

        :return: numpy array with the predicted values
        """

    @abc.abstractmethod
    def train_val_loop(self, X_train: np.array, y_train: np.array, X_val: np.array, y_val: np.array) -> np.array:
        """
        Method that runs the whole training and validation loop

        :param X_train: feature matrix for the training
        :param y_train: target vector for training
        :param X_val: feature matrix for validation
        :param y_val: target vector for validation

        :return: predictions on validation set
        """

    # General methods #
    def suggest_hyperparam_to_optuna(self, hyperparam_name: str):
        """
        Suggest a hyperparameter of hyperparam_dict to the optuna trial to optimize it.

        If you want to add a parameter to your model / in your pipeline to be optimized, you need to call this method

        :param hyperparam_name: name of the hyperparameter to be tuned (see :obj:`~easypheno.model._base_model.BaseModel.define_hyperparams_to_tune`)

        :return: suggested value
        """
        # Get specification of the hyperparameter
        if hyperparam_name in self.all_hyperparams:
            spec = self.all_hyperparams[hyperparam_name]
        else:
            raise Exception(hyperparam_name + ' not found in all_hyperparams dictionary.')

        # Check if the hyperparameter already exists in the trial and needs a suffix
        # (e.g. same dropout specification for multiple layers that should be optimized individually)
        if hyperparam_name in self.optuna_trial.params:
            counter = 1
            while True:
                current_name = hyperparam_name + '_' + str(counter)
                if current_name not in self.optuna_trial.params:
                    optuna_param_name = current_name
                    break
                counter += 1
        else:
            optuna_param_name = hyperparam_name

        # Read dict with specification for the hyperparamater and suggest it to the trial
        if spec['datatype'] == 'categorical':
            if 'list_of_values' not in spec:
                raise Exception(
                    '"list of values" for ' + hyperparam_name + ' not in hyperparams_dict. '
                                                                'Check define_hyperparams_to_tune() of the model.'
                )
            suggested_value = \
                self.optuna_trial.suggest_categorical(name=optuna_param_name, choices=spec['list_of_values'])
        elif spec['datatype'] in ['float', 'int']:
            if 'step' in spec:
                step = spec['step']
            else:
                step = None if spec['datatype'] == 'float' else 1
            log = spec['log'] if 'log' in spec else False
            if 'lower_bound' not in spec or 'upper_bound' not in spec:
                raise Exception(
                    '"lower_bound" or "upper_bound" for ' + hyperparam_name + ' not in all_hyperparams. '
                    'Check define_hyperparams_to_tune() of the model.'
                )
            if spec['datatype'] == 'int':
                suggested_value = self.optuna_trial.suggest_int(
                    name=optuna_param_name, low=spec['lower_bound'], high=spec['upper_bound'], step=step, log=log
                )
            else:
                suggested_value = self.optuna_trial.suggest_float(
                    name=optuna_param_name, low=spec['lower_bound'], high=spec['upper_bound'], step=step, log=log
                )
        else:
            raise Exception(
                spec['datatype'] + ' is not a valid parameter. Check define_hyperparams_to_tune() of the model.'
            )
        return suggested_value

    def suggest_all_hyperparams_to_optuna(self) -> dict:
        """
        Some models accept a dictionary with the model parameters.
        This method suggests all hyperparameters in all_hyperparams and gives back a dictionary containing them.

        :return: dictionary with suggested hyperparameters
        """
        for param_name in self.all_hyperparams.keys():
            _ = self.suggest_hyperparam_to_optuna(param_name)
        return self.optuna_trial.params

    def save_model(self, path: pathlib.Path, filename: str):
        """
        Persist the whole model object on a hard drive
        (can be loaded with :obj:`~easypheno.model._model_functions.load_model`)

        :param path: path where the model will be saved
        :param filename: filename of the model
        """
        joblib.dump(self, path.joinpath(filename), compress=3)



import torch

from . import _torch_model


class Cnn(_torch_model.TorchModel):
    """
    Implementation of a class for a Convolutional Neural Network (CNN).

    See :obj:`~easypheno.model._base_model.BaseModel` and :obj:`~easypheno.model._torch_model.TorchModel` for more information on the attributes.
    """
    standard_encoding = 'onehot'
    possible_encodings = ['onehot']

    def define_model(self) -> torch.nn.Sequential:
        """
        Definition of a CNN network.

        Architecture:

            - N_LAYERS of (Conv1d + BatchNorm + Dropout)
            - MaxPool1d, Flatten, Linear, BatchNorm, Dropout
            - Linear output layer

        Kernel sizes for convolutional and max pooling layers may be fixed or optimized.
        Same applies for strides, number of output channels of the first convolutional layer, dropout rate,
        frequency of a doubling of the output channels and number of units in the first linear layer.
        """
        n_layers = self.suggest_hyperparam_to_optuna('n_layers')
        model = []
        act_function = self.get_torch_object_for_string(string_to_get=self.suggest_hyperparam_to_optuna('act_function'))
        in_channels = self.width_onehot
        kernel_size = self.suggest_hyperparam_to_optuna('kernel_size')
        stride = max(1, int(kernel_size * self.suggest_hyperparam_to_optuna('stride_perc_of_kernel_size')))
        out_channels = 2 ** 2  # self.suggest_hyperparam_to_optuna('initial_out_channels_exp')
        frequency_out_channels_doubling = 2  # self.suggest_hyperparam_to_optuna('frequency_out_channels_doubling')
        p = self.suggest_hyperparam_to_optuna('dropout')
        for layer in range(n_layers):
            model.append(torch.nn.Conv1d(in_channels=in_channels, out_channels=out_channels,
                                         kernel_size=kernel_size, stride=stride))
            model.append(act_function)
            model.append(torch.nn.BatchNorm1d(num_features=out_channels))
            model.append(torch.nn.Dropout(p))
            in_channels = out_channels
            if ((layer+1) % frequency_out_channels_doubling) == 0:
                out_channels *= 2
        model.append(torch.nn.MaxPool1d(kernel_size=kernel_size))
        model.append(torch.nn.Flatten())
        in_features = torch.nn.Sequential(*model)(torch.zeros(size=(1, self.width_onehot, self.n_features))).shape[1]
        out_features = int(in_features * self.suggest_hyperparam_to_optuna('n_units_factor_linear_layer'))
        model.append(torch.nn.Linear(in_features=in_features, out_features=out_features))
        model.append(act_function)
        model.append(torch.nn.BatchNorm1d(num_features=out_features))
        model.append(torch.nn.Dropout(p))
        model.append(torch.nn.Linear(in_features=out_features, out_features=self.n_outputs))
        return torch.nn.Sequential(*model)

    def define_hyperparams_to_tune(self) -> dict:
        """
        See :obj:`~easypheno.model._base_model.BaseModel` for more information on the format.

        See :obj:`~easypheno.model._torch_model.TorchModel` for more information on hyperparameters common for all torch models.
        """
        kernel_size = {
            'datatype': 'int',
            'lower_bound': 2,
            'upper_bound': 8
        }
        n_units_factor_linear_layer = {
            # Number of units in the linear layer after flattening in relation to the number of inputs
            'datatype': 'float',
            'lower_bound': 0.2,
            'upper_bound': 1,
            'step': 0.05
        }

        if self.n_features > 15000:
            kernel_size = {
                'datatype': 'int',
                'lower_bound': 4,
                'upper_bound': 10,
                'step': 2
            }
            n_units_factor_linear_layer = {
                # Number of units in the linear layer after flattening in relation to the number of inputs
                'datatype': 'float',
                'lower_bound': 0.2,
                'upper_bound': 0.5,
                'step': 0.05
            }
        if self.n_features > 50000:
            kernel_size = {
                'datatype': 'int',
                'lower_bound': 8,
                'upper_bound': 14,
                'step': 2
            }
        if self.n_features > 100000:
            kernel_size = {
                'datatype': 'int',
                'lower_bound': 10,
                'upper_bound': 25,
                'step': 5
            }
        return {
            'n_layers': {
                'datatype': 'int',
                'lower_bound': 1,
                'upper_bound': 3
            },
            'initial_out_channels_exp': {
                # Exponent with base 2 to get number of output channels of the first convolutional layer
                'datatype': 'int',
                'lower_bound': 1,
                'upper_bound': 2
            },
            'frequency_out_channels_doubling': {
                # Frequency of doubling the initial output channels after a certain number of layers
                'datatype': 'int',
                'lower_bound': 1,
                'upper_bound': 2
            },
            'kernel_size': kernel_size,
            'stride_perc_of_kernel_size': {
                # Stride in relation to the kernel size
                'datatype': 'float',
                'lower_bound': 0.5,
                'upper_bound': 1,
                'step': 0.1
            },
            'n_units_factor_linear_layer': n_units_factor_linear_layer
        }



import abc
import numpy as np

from . import _base_model


class SklearnModel(_base_model.BaseModel, abc.ABC):
    """
    Parent class based on :obj:`~easypheno.model._base_model.BaseModel` for all models with a sklearn-like API to share
    functionalities. See :obj:`~easypheno.model._base_model.BaseModel` for more information.

    **Attributes**

        *Inherited attributes*

        See :obj:`~easypheno.model._base_model.BaseModel`
    """

    def retrain(self, X_retrain: np.array, y_retrain: np.array):
        """
        Implementation of the retraining for models with sklearn-like API.
        See :obj:`~easypheno.model._base_model.BaseModel` for more information.
        """
        self.model.fit(X_retrain, np.ravel(y_retrain))

    def predict(self, X_in: np.array) -> np.array:
        """
        Implementation of a prediction based on input features for models with sklearn-like API.
        See :obj:`~easypheno.model._base_model.BaseModel` for more information.
        """
        return np.reshape(self.model.predict(X_in), (-1, 1))

    def train_val_loop(self, X_train: np.array, y_train: np.array, X_val: np.array, y_val: np.array) -> np.array:
        """
        Implementation of a train and validation loop for models with sklearn-like API.
        See :obj:`~easypheno.model._base_model.BaseModel` for more information.
        """
        self.model.fit(X_train, np.ravel(y_train))
        return self.predict(X_in=X_val)



import argparse

from.results_analysis_synthetic_data import featimps_vs_simulation

if __name__ == "__main__":
    """
    Run file to generate statistics and plots of feature importances versus effect sizes on synthetic data
    """
    parser = argparse.ArgumentParser()
    parser.add_argument("-rd", "--results_dir", type=str,
                        help="Provide the full path of the directory where your results are stored "
                             "(name of the genotype matrix level)")
    parser.add_argument("-simd", "--sim_config_dir", type=str,
                        help="Provide the full path of the directory where the simulation configs are stored")
    parser.add_argument("-sd", "--save_dir", type=str, default=None,
                        help="Define save directory for the plots. Default is the same as results directory.")

    args = vars(parser.parse_args())
    results_directory_genotype_level = args['results_dir']
    sim_config_dir = args['sim_config_dir']
    save_dir = args['save_dir']

    featimps_vs_simulation(
        results_directory_genotype_level=results_directory_genotype_level,
        sim_config_dir=sim_config_dir, save_dir=save_dir
    )



import numpy as np
import rpy2
from rpy2.robjects import numpy2ri
rpy2.robjects.numpy2ri.activate()
import rpy2.robjects as robjects
from rpy2.robjects.packages import importr
from . import _param_free_base_model


class Bayes_R(_param_free_base_model.ParamFreeBaseModel):
    """
    Implementation of a class for Bayesian alphabet.

    *Attributes*

        *Inherited attributes*

        See :obj:`~easypheno.model._param_free_base_model.ParamFreeBaseModel` for more information on the attributes.

        *Additional attributes*

        - mu (*np.array*): intercept
        - beta (*np.array*): effect size
        - model_name (*str*): model to use (BayesA, BayesB or BayesC)
        - n_iter (*int*): iterations for sampling
        - burn_in (*int*): warmup/burnin for sampling
    """
    standard_encoding = '012'
    possible_encodings = ['101']

    def __init__(self, task: str, model_name: str, encoding: str = None, n_iter: int = 6000, burn_in: int = 1000):
        super().__init__(task=task, encoding=encoding)
        self.model_name = model_name
        self.n_iter = n_iter
        self.burn_in = burn_in
        self.mu = None
        self.beta = None

    def fit(self, X: np.array, y: np.array) -> np.array:
        """
        Implementation of fit function for Bayesian alphabet imported from R.

        See :obj:`~easypheno.model._param_free_base_model.ParamFreeBaseModel` for more information.
        """
        # import necessary R packages
        base = importr('base')
        BGLR = importr('BGLR')

        # create R objects for X and y
        R_X = robjects.r['matrix'](X, nrow=X.shape[0], ncol=X.shape[1])
        R_y = robjects.FloatVector(y)

        # run BGLR for BayesB
        ETA = base.list(base.list(X=R_X, model=self.model_name))
        fmBB = BGLR.BGLR(y=R_y, ETA=ETA, verbose=True, nIter=self.n_iter, burnIn=self.burn_in)

        # save results as numpy arrays
        self.beta = np.asarray(fmBB.rx2('ETA').rx2(1).rx2('b'))
        self.mu = fmBB.rx2('mu')
        return self.predict(X_in=X)

    def predict(self, X_in: np.array) -> np.array:
        """
        Implementation of predict function for Bayesian alphabet model imported from R.

        See :obj:`~easypheno.model._param_free_base_model.ParamFreeBaseModel` for more information.
        """
        return self.mu + np.matmul(X_in, self.beta)



import abc
import numpy as np
import optuna
import tensorflow as tf
import joblib
import pathlib

from . import _base_model


class TensorflowModel(_base_model.BaseModel, abc.ABC):
    """
    Parent class based on :obj:`~easypheno.model._base_model.BaseModel` for all TensorFlow models to share functionalities.
    See :obj:`~easypheno.model._base_model.BaseModel` for more information.

    **Attributes**

        *Inherited attributes*

        See :obj:`~easypheno.model._base_model.BaseModel`.

        *Additional attributes*

        - n_features (*int*): Number of input features to the model
        - width_onehot (*int*): Number of input channels in case of onehot encoding
        - batch_size (*int*): Batch size for batch-based training
        - n_epochs (*int*): Number of epochs for optimization
        - optimizer (*tf.keras.optimizers.Optimizer*): optimizer for model fitting
        - loss_fn: loss function for model fitting
        - early_stopping_patience (*int*): epochs without improvement before early stopping
        - early_stopping_point (*int*): epoch at which early stopping occured
        - early_stopping_callback (*tf.keras.callbacks.EarlyStopping*): callback for early stopping

    :param task: ML task (regression or classification) depending on target variable
    :param optuna_trial: optuna.trial.Trial : trial of optuna for optimization
    :param encoding: the encoding to use (standard encoding or user-defined)
    :param n_features: Number of input features to the model
    :param width_onehot: Number of input channels in case of onehot encoding
    :param batch_size: Batch size for batch-based training
    :param n_epochs: Number of epochs for optimization
    :param early_stopping_point: Stop training at defined epoch
    """

    def __init__(self, task: str, optuna_trial: optuna.trial.Trial, encoding: str = None, n_outputs: int = 1,
                 n_features: int = None, width_onehot: int = None, batch_size: int = None, n_epochs: int = None,
                 early_stopping_point: int = None):
        self.all_hyperparams = self.common_hyperparams()  # add hyperparameters commonly optimized for all torch models
        self.n_features = n_features
        self.width_onehot = width_onehot  # relevant for models using onehot encoding e.g. CNNs
        super().__init__(task=task, optuna_trial=optuna_trial, encoding=encoding, n_outputs=n_outputs)
        self.batch_size = \
            batch_size if batch_size is not None else self.suggest_hyperparam_to_optuna('batch_size')
        self.n_epochs = n_epochs if n_epochs is not None else self.suggest_hyperparam_to_optuna('n_epochs')
        self.optimizer = tf.keras.optimizers.Adam(learning_rate=self.suggest_hyperparam_to_optuna('learning_rate'))
        self.loss_fn = tf.keras.losses.SparseCategoricalCrossentropy(from_logits=True) if task == 'classification' \
            else tf.keras.losses.MeanSquaredError()
        # early stopping if there is no improvement on validation loss for a certain number of epochs
        self.early_stopping_patience = self.suggest_hyperparam_to_optuna('early_stopping_patience')
        self.early_stopping_point = early_stopping_point
        self.early_stopping_callback = tf.keras.callbacks.EarlyStopping(
            monitor='val_loss', patience=self.early_stopping_patience, mode='min', restore_best_weights=True,
            min_delta=0.1
        )
        self.model.compile(self.optimizer, loss=self.loss_fn)

    def train_val_loop(self, X_train: np.array, y_train: np.array, X_val: np.array, y_val: np.array) -> np.array:
        """
        Implementation of a train and validation loop for  TensorFlow models.
        See :obj:`~easypheno.model._base_model.BaseModel` for more information
        """
        history = self.model.fit(x=X_train, y=y_train, batch_size=self.batch_size, epochs=self.n_epochs,
                                 validation_data=(X_val, y_val), validation_freq=1, verbose=2,
                                 callbacks=[self.early_stopping_callback])
        if len(history.history['loss']) != self.n_epochs:
            self.early_stopping_point = len(history.history['loss']) - self.early_stopping_patience
            print("Early Stopping at " + str(self.early_stopping_point + self.early_stopping_patience)
                  + ' of ' + str(self.n_epochs))
        return self.predict(X_in=X_val)

    def retrain(self, X_retrain: np.array, y_retrain: np.array):
        """
        Implementation of the retraining for PyTorch models.
        See :obj:`~easypheno.model._base_model.BaseModel` for more information
        """
        n_epochs_to_retrain = self.n_epochs if self.early_stopping_point is None else self.early_stopping_point
        self.model.fit(x=X_retrain, y=y_retrain, batch_size=self.batch_size, epochs=n_epochs_to_retrain, verbose=2)

    def predict(self, X_in: np.array) -> np.array:
        """
        Implementation of a prediction based on input features for PyTorch models.
        See :obj:`~easypheno.model._base_model.BaseModel` for more information
        """
        dataloader = self.get_dataloader(X=X_in, shuffle=False)
        predictions = None
        for inputs in dataloader:
            outputs = self.model(inputs, training=False)
            # concat
            predictions = outputs.numpy() if predictions is None else np.vstack((predictions, outputs.numpy()))
        if self.task == 'classification':
            predictions = predictions.argmax(axis=1)
        return predictions

    def get_dataloader(self, X: np.array, y: np.array = None, shuffle: bool = True) -> tf.data.Dataset:
        """
        Get a dataloader using the specified data and batch size

        :param X: feature matrix to use
        :param y: optional target vector to use
        :param shuffle: shuffle parameter for DataLoader

        :return: batched dataset
        """
        # drop last sample if last batch would only contain one sample
        if (len(X) % self.batch_size) == 1:
            X = X[:-1]
            y = y[:-1] if y is not None else None
        dataset = tf.data.Dataset.from_tensor_slices((X, y)) if y is not None \
            else tf.data.Dataset.from_tensor_slices(X)
        if shuffle:
            dataset = dataset.shuffle(buffer_size=len(dataset))
        return dataset.batch(batch_size=self.batch_size)

    @staticmethod
    def common_hyperparams():
        """
        Add hyperparameters that are common for PyTorch models.
        Do not need to be included in optimization for every child model.
        Also See :obj:`~easypheno.model._base_model.BaseModel` for more information
        """
        return {
            'dropout': {
                'datatype': 'float',
                'lower_bound': 0,
                'upper_bound': 0.5,
                'step': 0.1
            },
            'act_function': {
                'datatype': 'categorical',
                'list_of_values': ['relu']
            },
            'batch_size': {
                'datatype': 'categorical',
                'list_of_values': [4, 8, 16, 32, 64]
            },
            'n_epochs': {
                'datatype': 'categorical',
                'list_of_values': [100, 500, 1000, 10000]
            },
            'learning_rate': {
                'datatype': 'categorical',
                'list_of_values': [1e-6, 1e-5, 1e-4, 1e-3, 1e-2, 1e-1]
            },
            'early_stopping_patience': {
                'datatype': 'int',
                'lower_bound': 0,
                'upper_bound': 10,
                'step': 10
            },
            'l1_factor': {
                'datatype': 'float',
                'lower_bound': 0,
                'upper_bound': 10**3
            }
        }

    def save_model(self, path: pathlib.Path, filename: str):
        """
        Method to persist the whole model object on a hard drive (can be loaded with :obj:`~easypheno.model._model_functions.load_model`)

        :param path: path where the model will be saved
        :param filename: filename of the model
        """
        optimizer = self.optimizer
        # special case for serialization of optimizer prior to saving
        self.optimizer = tf.keras.optimizers.serialize(self.optimizer)
        joblib.dump(self, path.joinpath(filename), compress=3)
        self.optimizer = optimizer



import pandas as pd
import numpy as np
import h5py
import pathlib
from sklearn.model_selection import train_test_split
from sklearn.model_selection import StratifiedKFold
from pandas_plink import read_plink1_bin

from ..utils import helper_functions
from . import encoding_functions as enc


def prepare_data_files(data_dir: pathlib.Path, genotype_matrix_name: str, phenotype_matrix_name: str, phenotype: str,
                       datasplit: str, n_outerfolds: int, n_innerfolds: int, test_set_size_percentage: int,
                       val_set_size_percentage: int, models, user_encoding: str, maf_percentage: int):
    """
    Prepare all data files for a common format: genotype matrix, phenotype matrix and index file.

    First check if genotype file is .h5 file (standard format of this framework):

    - YES: First check if all required information is present in the file, raise Exception if not. Then check if index file exists:

        - NO: Load genotype and create all required index files
        - YES: Append all required data splits and maf-filters to index file

    - NO: Load genotype and create all required files

    :param data_dir: data directory where the phenotype and genotype matrix are stored
    :param genotype_matrix_name: name of the genotype matrix including datatype ending
    :param phenotype_matrix_name: name of the phenotype matrix including datatype ending
    :param phenotype: name of the phenotype to predict
    :param datasplit: datasplit to use. Options are: nested-cv, cv-test, train-val-test
    :param n_outerfolds: number of outerfolds relevant for nested-cv
    :param n_innerfolds: number of folds relevant for nested-cv and cv-test
    :param test_set_size_percentage: size of the test set relevant for cv-test and train-val-test
    :param val_set_size_percentage: size of the validation set relevant for train-val-test
    :param models: models to consider
    :param user_encoding: encoding specified by the user
    :param maf_percentage: threshold for MAF filter as percentage value
    """
    print('Check if all data files have the required format')
    if data_dir.joinpath(genotype_matrix_name).with_suffix('.h5').is_file() and \
            (genotype_matrix_name.split('.')[-1] != 'h5'):
        print("Found same file name with ending .h5")
        print("Assuming that the raw file was already prepared using our pipepline. Will continue with the .h5 file.")
        genotype_matrix_name = genotype_matrix_name.split('.')[0] + '.h5'
    suffix = genotype_matrix_name.split('.')[-1]
    if suffix in ('h5', 'hdf5', 'h5py'):
        # Genotype matrix has standard file format -> check information in the file
        check_genotype_h5_file(data_dir=data_dir, genotype_matrix_name=genotype_matrix_name,
                               encodings=enc.get_encoding(models=models, user_encoding=user_encoding))
        print('Genotype file available in required format, check index file now.')
        # Check / create index files
        if check_index_file(data_dir=data_dir, genotype_matrix_name=genotype_matrix_name,
                            phenotype_matrix_name=phenotype_matrix_name, phenotype=phenotype):
            print('Index file ' + genotype_matrix_name.split('.')[0] + '-'
                    + phenotype_matrix_name.split('.')[0] + '-' + phenotype + '.h5' + ' already exists.'
                    ' Will append required filters and data splits now.')
            append_index_file(data_dir=data_dir, genotype_matrix_name=genotype_matrix_name,
                              phenotype_matrix_name=phenotype_matrix_name, phenotype=phenotype,
                              maf_percentage=maf_percentage,
                              datasplit=datasplit, n_outerfolds=n_outerfolds, n_innerfolds=n_innerfolds,
                              test_set_size_percentage=test_set_size_percentage,
                              val_set_size_percentage=val_set_size_percentage)
            print('Done checking data files. All required datasets are available.')
        else:
            print('Index file ' + genotype_matrix_name.split('.')[0] + '-' + phenotype_matrix_name.split('.')[0]
                  + '-' + phenotype + '.h5' + ' does not fulfill requirements. '
                                              'Will load genotype and phenotype matrix and create new index file.')
            save_all_data_files(data_dir=data_dir, genotype_matrix_name=genotype_matrix_name,
                                phenotype_matrix_name=phenotype_matrix_name, phenotype=phenotype,
                                models=models, user_encoding=user_encoding, maf_percentage=maf_percentage,
                                datasplit=datasplit, n_outerfolds=n_outerfolds, n_innerfolds=n_innerfolds,
                                test_set_size_percentage=test_set_size_percentage,
                                val_set_size_percentage=val_set_size_percentage)
            print('Done checking data files. All required datasets are available.')
    else:
        print('Genotype file not in required format. Will load genotype matrix and save as .h5 file. Will also create '
              'required index file.')
        save_all_data_files(data_dir=data_dir, genotype_matrix_name=genotype_matrix_name,
                            phenotype_matrix_name=phenotype_matrix_name, phenotype=phenotype,
                            models=models, user_encoding=user_encoding, maf_percentage=maf_percentage,
                            datasplit=datasplit, n_outerfolds=n_outerfolds, n_innerfolds=n_innerfolds,
                            test_set_size_percentage=test_set_size_percentage,
                            val_set_size_percentage=val_set_size_percentage)
        print('Done checking data files. All required datasets are available.')


def check_genotype_h5_file(data_dir: pathlib.Path, genotype_matrix_name: str, encodings: list):
    """
    Check .h5 genotype file. Should contain:

    - sample_ids: vector with sample names of genotype matrix,
    - snp_ids: vector with SNP identifiers of genotype matrix,
    - X_{enc}: (samples x SNPs)-genotype matrix in enc encoding, where enc might refer to:

        - '012': additive (number of minor alleles)
        - 'raw': raw (alleles)

    :param data_dir: data directory where the phenotype and genotype matrix are stored
    :param genotype_matrix_name: name of the phenotype matrix including datatype ending
    :param encodings: list of needed encodings
    """
    with h5py.File(data_dir.joinpath(genotype_matrix_name), "r") as f:
        keys = list(f.keys())
        if {'sample_ids', 'snp_ids'}.issubset(keys):
            # check if required encoding is available or can be created
            for elem in encodings:
                if f'X_{elem}' not in f and f'X_{enc.get_base_encoding(encoding=elem)}' not in f:
                    raise Exception('Genotype in ' + elem + ' encoding missing. Can not create required encoding. '
                                                            'See documentation for help')
        else:
            raise Exception('sample_ids and/or snp_ids are missing in' + genotype_matrix_name)


def check_index_file(data_dir: pathlib.Path, genotype_matrix_name: str, phenotype_matrix_name: str, phenotype: str) -> bool:
    """
    Check if index file is available and if the datasets 'y', 'matched_sample_ids', 'X_index', 'y_index' and
    'ma_frequency' exist.

    :param data_dir: data directory where the phenotype and genotype matrix are stored
    :param genotype_matrix_name: name of the genotype matrix including datatype ending
    :param phenotype_matrix_name: name of the phenotype matrix including datatype ending
    :param phenotype: name of the phenotype to predict

    :return: bool reflecting check result
    """
    index_file = data_dir.joinpath(genotype_matrix_name.split('.')[0] + '-' + phenotype_matrix_name.split('.')[0] \
                 + '-' + phenotype + '.h5')
    if index_file.is_file():
        matched_datasets = ['y', 'matched_sample_ids', 'X_index', 'y_index', 'non_informative_filter', 'ma_frequency']
        with h5py.File(index_file, 'a') as f:
            if 'matched_data' in f and all(z in f['matched_data'] for z in matched_datasets):
                return True
            else:
                return False
    else:
        return False


def save_all_data_files(data_dir: pathlib.Path, genotype_matrix_name: str, phenotype_matrix_name: str, phenotype: str,
                        models, user_encoding: str, maf_percentage: int,
                        datasplit: str, n_outerfolds: int, n_innerfolds: int,
                        test_set_size_percentage: int, val_set_size_percentage: int):
    """
    Prepare and save all required data files:

    - genotype matrix in unified format as .h5 file with,
    - phenotype matrix in unified format as .csv file,
    - file containing maf filter and data split indices as .h5

    :param data_dir: data directory where the phenotype and genotype matrix are stored
    :param genotype_matrix_name: name of the genotype matrix including datatype ending
    :param phenotype_matrix_name: name of the phenotype matrix including datatype ending
    :param phenotype: name of the phenotype to predict
    :param models: models to consider
    :param user_encoding: encoding specified by the user
    :param maf_percentage: threshold for MAF filter as percentage value
    :param datasplit: datasplit to use. Options are: nested-cv, cv-test, train-val-test
    :param n_outerfolds: number of outerfolds relevant for nested-cv
    :param n_innerfolds: number of folds relevant for nested-cv and cv-test
    :param test_set_size_percentage: size of the test set relevant for cv-test and train-val-test
    :param val_set_size_percentage: size of the validation set relevant for train-val-test
    """
    print('Load genotype file ' + str(data_dir.joinpath(genotype_matrix_name)))
    X, X_ids, _ = check_transform_format_genotype_matrix(data_dir=data_dir, genotype_matrix_name=genotype_matrix_name,
                                                      models=models, user_encoding=user_encoding)
    print('Have genotype matrix. Load phenotype ' + phenotype + ' from ' + str(data_dir.joinpath(phenotype_matrix_name)))
    y = check_and_load_phenotype_matrix(data_dir=data_dir,
                                        phenotype_matrix_name=phenotype_matrix_name, phenotype=phenotype)
    print('Have phenotype vector. Start matching genotype and phenotype.')
    X, y, sample_ids, X_index, y_index = genotype_phenotype_matching(X=X, X_ids=X_ids, y=y)
    print('Done matching genotype and phenotype. Create index file now.')
    create_index_file(data_dir=data_dir, genotype_matrix_name=genotype_matrix_name,
                      phenotype_matrix_name=phenotype_matrix_name, phenotype=phenotype,
                      datasplit=datasplit, n_outerfolds=n_outerfolds, n_innerfolds=n_innerfolds,
                      test_set_size_percentage=test_set_size_percentage,
                      val_set_size_percentage=val_set_size_percentage,
                      maf_percentage=maf_percentage, X=X, y=y, sample_ids=sample_ids, X_index=X_index, y_index=y_index
                      )


def check_transform_format_genotype_matrix(data_dir: pathlib.Path, genotype_matrix_name: str, models,
                                           user_encoding: str, save_h5: bool = True) -> (np.array, np.array, np.array):
    """
    Check the format of the specified genotype matrix.

    Unified genotype matrix will be saved in subdirectory data and named NAME_OF_GENOTYPE_MATRIX.h5

    Unified format of the .h5 file of the genotype matrix required for the further processes:

    - mandatory:

        - sample_ids: vector with sample names of genotype matrix,
        - SNP_ids: vector with SNP identifiers of genotype matrix,
        - X_{enc}: (samples x SNPs)-genotype matrix in enc encoding, where enc might refer to:

            - '012': additive (number of minor alleles)
            - 'raw': raw  (alleles)

    - optional: genotype in additional encodings

    Accepts .h5, .hdf5, .h5py, .csv, PLINK binary and PLINK files. .h5, .hdf5, .h5py files must satisfy the unified
    format. If the genotype matrix contains constant SNPs, those will be removed and a new file will be saved.
    Will open .csv, PLINK and binary PLINK files and generate required .h5 format.

    :param data_dir: data directory where the phenotype and genotype matrix are stored
    :param genotype_matrix_name: name of the genotype matrix including datatype ending
    :param models: models to consider
    :param user_encoding: encoding specified by the user
    :param save_h5: save genotype in unified h5 format if True, default is True

    :return: genotype matrix (raw encoded if present, 012 encoded otherwise), sample ids and SNP ids
    """
    genotype_file = data_dir.joinpath(genotype_matrix_name)
    suffix = genotype_file.suffix
    encoding = enc.get_encoding(models=models, user_encoding=user_encoding)
    if suffix in ('.h5', '.hdf5', '.h5py'):
        with h5py.File(genotype_file, "r") as f:
            sample_ids = f['sample_ids'][:].astype(str)
            snp_ids = f['snp_ids'][:].astype(str)
            if 'X_raw' in f:
                X = f['X_raw'][:]
            elif 'X_012' in f:
                X = f['X_012'][:]
    else:
        if suffix == '.csv':
            sample_ids, snp_ids, X = check_genotype_csv_file(data_dir=data_dir,
                                                             genotype_matrix_name=genotype_matrix_name,
                                                             encodings=encoding)

        elif suffix in ('.bed', '.bim', '.fam'):
            sample_ids, snp_ids, X = check_genotype_binary_plink_file(data_dir=data_dir,
                                                                      genotype_matrix_name=genotype_matrix_name)

        elif suffix in ('.map', '.ped'):
            sample_ids, snp_ids, X = check_genotype_plink_file(data_dir=data_dir,
                                                               genotype_matrix_name=genotype_matrix_name)
        else:
            raise Exception('Only accept .h5, .hdf5, .h5py, .csv, binary PLINK and PLINK genotype files. '
                            'See documentation for help.')
        if save_h5:
            create_genotype_h5_file(data_dir=data_dir, genotype_matrix_name=genotype_matrix_name,
                                    sample_ids=sample_ids, snp_ids=snp_ids, X=X)
    # perform sanity checks
    check_genotype_shape(X=X, sample_ids=sample_ids, snp_ids=snp_ids)
    if check_duplicate_samples(sample_ids=sample_ids):
        raise Exception('The genotype matrix contains duplicate samples. Please check again.')
    return X, sample_ids, snp_ids


def check_genotype_csv_file(data_dir: pathlib.Path, genotype_matrix_name: str, encodings: list) \
        -> (np.array, np.array, np.array):
    """
    Load .csv genotype file. File must have the following structure:
    First column must contain the sample ids, the column names should be the SNP ids.
    The values should be the genotype matrix either in additive encoding or in raw encoding.
    If the name of the first column is 'MarkerID' it is assumed that the rows contain the markers and the column contain
    the samples and the genotype matrix will be transposed.
    If the csv file contains the genotype in biallelic notation (i.e. 'AA', 'AT', ...), this function generates a
    genotype matrix in iupac notation (i.e. 'A', 'W', ...).

    :param data_dir: data directory where the phenotype and genotype matrix are stored
    :param genotype_matrix_name: name of the genotype matrix including datatype ending
    :param encodings: list of needed encodings

    :return: sample ids, SNP ids and genotype in additive / raw encoding
    """
    gt = pd.read_csv(data_dir.joinpath(genotype_matrix_name), index_col=0)
    if gt.index.name == 'MarkerID':
        gt = gt.T
    snp_ids = np.asarray(gt.columns.values)
    sample_ids = np.asarray(gt.index)
    X = np.asarray(gt.values)
    enc_of_X = enc.check_encoding_of_genotype(X=X)
    # if genotype in biallelic notation, will change to iupac notation
    if enc_of_X == 'biallelic':
        iupac_map = {"AA": "A", "GG": "G", "TT": "T", "CC": "C", "AG": "R", "GA": "R", "CT": "Y", "TC": "Y", "GC": "S",
                     "CG": "S", "AT": "W", "TA": "W", "GT": "K", "TG": "K", "AC": "M", "CA": "M"}
        gt = gt.applymap(lambda x: x.decode() if isinstance(x, bytes) else x)
        X = np.asarray(gt.replace(iupac_map).values)
        enc_of_X = 'raw'
    # check encoding of X, only accept additive or raw and check if required encoding can be created
    for elem in encodings:
        if elem != enc_of_X and enc.get_base_encoding(encoding=elem) != enc_of_X:
            raise Exception('Genotype in ' + genotype_matrix_name + ' in wrong encoding. Can not create'
                            ' required encoding. See documentation for help.')

    return sample_ids, snp_ids, X


def check_genotype_binary_plink_file(data_dir: pathlib.Path, genotype_matrix_name: str) -> (np.array, np.array, np.array):
    """
    Load binary PLINK file, .bim, .fam, .bed files with same prefix need to be in same folder.
    Compute additive and raw encoding of genotype

    :param data_dir: data directory where the phenotype and genotype matrix are stored
    :param genotype_matrix_name: name of the genotype matrix including datatype ending

    :return: sample ids, SNP ids and genotype in raw encoding
    """
    gt_file = data_dir.joinpath(genotype_matrix_name).with_suffix('').as_posix()
    gt = read_plink1_bin(gt_file + '.bed', gt_file + '.bim', gt_file + '.fam', ref="a0", verbose=False)
    sample_ids = np.array(gt['fid'], dtype=str).flatten()
    snp_ids = np.array(gt['snp']).flatten()
    # get raw encoding
    iupac_map = {b"AA": "A", b"GG": "G", b"TT": "T", b"CC": "C", b"AG": "R", b"GA": "R", b"CT": "Y", b"TC": "Y",
                 b"GC": "S", b"CG": "S", b"AT": "W", b"TA": "W", b"GT": "K", b"TG": "K", b"AC": "M", b"CA": "M"}
    a_0 = np.array(gt.a1.values, dtype='S1')
    a_2 = np.array(gt.a0.values, dtype='S1')
    a_1 = []
    for i in range(len(a_0)):
        a_1.append(iupac_map[a_0[i] + a_2[i]])
    a = np.stack((a_0, np.array(a_1), a_2))
    col = np.arange(a.shape[1])
    X_012 = np.array(gt.values)
    X_raw = a[X_012.astype(int), col]
    return sample_ids, snp_ids, X_raw


def check_genotype_plink_file(data_dir: pathlib.Path, genotype_matrix_name: str) -> (np.array, np.array, np.array):
    """
    Load PLINK files, .map and .ped file with same prefix need to be in same folder.
    Accepts GENOTYPENAME.ped and GENOTYPENAME.map as input

    :param data_dir: data directory where the phenotype and genotype matrix are stored
    :param genotype_matrix_name: name of the genotype matrix including datatype ending

    :return: sample ids, SNP ids and genotype in raw encoding
    """
    gt_file = data_dir.joinpath(genotype_matrix_name).with_suffix('').as_posix()
    with open(gt_file + '.map', 'r') as f:
        snp_ids = []
        for line in f:
            tmp = line.strip().split(" ")
            snp_ids.append(tmp[1].strip())
    snp_ids = np.array(snp_ids)
    iupac_map = {"AA": "A", "GG": "G", "TT": "T", "CC": "C", "AG": "R", "GA": "R", "CT": "Y", "TC": "Y", "GC": "S",
                 "CG": "S", "AT": "W", "TA": "W", "GT": "K", "TG": "K", "AC": "M", "CA": "M"}
    with open(gt_file + '.ped', 'r') as f:
        sample_ids = []
        X_raw = []
        for line in f:
            tmp = line.strip().split(" ")
            sample_ids.append(int(tmp[1].strip()))
            snps = []
            j = 6
            while j < len(tmp) - 1:
                snps.append(iupac_map[tmp[j] + tmp[j + 1]])
                j += 2
            X_raw.append(snps)
    sample_ids = np.array(sample_ids)
    X_raw = np.array(X_raw)
    return sample_ids, snp_ids, X_raw


def check_duplicate_samples(sample_ids: np.array) -> bool:
    """
    check if genotype matrix contain duplicate samples

    :param sample_ids: sample ids of genotype matrix

    :return: True if duplicates are present, False if not
    """
    return len(np.unique(sample_ids.flatten())) != len(sample_ids.flatten())


def check_genotype_shape(X: np.array, sample_ids: np.array, snp_ids: np.array):
    """
    Check if number of samples in sample_ids and genotype matrix match
    and if number of markers in snp_ids and genotype matrix match.

    :param X: genotype matrix
    :param sample_ids: vector containing sample ids of genotype
    :param snp_ids:  vector containing SNP ids of genotype
    """
    if X.shape[0] != sample_ids.shape[0]:
        raise Exception('Number of samples in genotype matrix and sample_ids do not match. Please check again.')
    if X.shape[1] != snp_ids.shape[0]:
        raise Exception('Number of markers in genotype matrix and snp_ids do not match. Please check again.')


def create_genotype_h5_file(data_dir: pathlib.Path, genotype_matrix_name: str,
                            sample_ids: np.array, snp_ids: np.array, X: np.array):
    """
    Save genotype matrix in unified .h5 file.

    Structure:

    - sample_ids
    - snp_ids
    - X_raw (or X_012 if X_raw not available)

    :param data_dir: data directory where the phenotype and genotype matrix are stored
    :param genotype_matrix_name: name of the genotype matrix including datatype ending
    :param sample_ids: array containing sample ids of genotype data
    :param snp_ids: array containing snp ids of genotype data
    :param X: matrix containing genotype either in raw or in additive encoding
    """
    x_file = data_dir.joinpath(genotype_matrix_name).with_suffix('.h5')
    print('Save unified genotype file ' + str(x_file))
    with h5py.File(x_file, 'w') as f:
        f.create_dataset('sample_ids', data=sample_ids.astype(bytes), chunks=True, compression="gzip")
        f.create_dataset('snp_ids', data=snp_ids.astype(bytes), chunks=True, compression="gzip")
        encoding = enc.check_encoding_of_genotype(X=X)
        if encoding == 'raw':
            f.create_dataset('X_raw', data=X.astype(bytes), chunks=True, compression="gzip", compression_opts=7)
        elif encoding == '012':
            f.create_dataset('X_012', data=X, chunks=True, compression="gzip", compression_opts=7)
        else:
            raise Exception('Genotype neither in raw or additive encoding. Cannot save .h5 genotype file.')


def check_and_load_phenotype_matrix(data_dir: pathlib.Path, phenotype_matrix_name: str, phenotype: str) -> pd.DataFrame:
    """
    Check and load the specified phenotype matrix. Only accept .csv, .pheno, .txt files.
    Sample ids need to be in first column, remaining columns should contain phenotypic values
    with phenotype name as column name

    :param data_dir: data directory where the phenotype and genotype matrix are stored
    :param phenotype_matrix_name: name of the phenotype matrix including datatype ending
    :param phenotype: name of the phenotype to predict

    :return: DataFrame with sample_ids as index and phenotype values as single column without NAN values
    """
    phenotype_file = data_dir.joinpath(phenotype_matrix_name)
    suffix = phenotype_file.suffix
    if suffix == ".csv":
        y = pd.read_csv(phenotype_file)
    elif suffix in (".pheno", ".txt"):
        y = pd.read_csv(phenotype_file, sep=" ")
    else:
        raise Exception('Only accept .csv, .pheno, .txt phenotype files. See documentation for help')
    y = y.sort_values(y.columns[0]).groupby(y.columns[0]).mean()
    if phenotype not in y.columns:
        raise Exception('Phenotype ' + phenotype + ' is not in phenotype file ' + phenotype_matrix_name +
                        ' See documentation for help')
    else:
        y = y[[phenotype]].dropna()
    return y


def genotype_phenotype_matching(X: np.array, X_ids: np.array, y: pd.DataFrame) -> tuple:
    """
    Match the handed over genotype and phenotype matrix for the phenotype specified by the user, i.e. compare sample ids

    :param X: genotype matrix in additive encoding
    :param X_ids: sample ids of genotype matrix
    :param y: pd.DataFrame containing sample ids of phenotype as index and phenotype values as single column

    :return: matched genotype matrix, matched sample ids, index arrays for genotype and phenotype to redo matching
    """
    y_ids = np.asarray(y.index, dtype=X_ids.dtype).flatten()
    (y_index, X_index) = (np.reshape(y_ids, (y_ids.shape[0], 1)) == X_ids).nonzero()
    if len(y_index) == 0:
        raise Exception('Samples of genotype and phenotype do not match.')
    X = get_matched_data(X, X_index)
    X_ids = get_matched_data(X_ids, X_index)
    y = get_matched_data(y.values.flatten(), y_index)
    return X, y, X_ids, X_index, y_index


def get_matched_data(data: np.array, index: np.array) -> np.array:
    """
    Get elements of data specified in index array

    :param data: matrix or array
    :param index: index array

    :return: data at selected indices
    """
    if data.ndim == 2:
        return data[index, :]
    else:
        return data[index]


def append_index_file(data_dir: pathlib.Path, genotype_matrix_name: str, phenotype_matrix_name: str, phenotype: str,
                      datasplit: str, n_outerfolds: int, n_innerfolds: int, test_set_size_percentage: int,
                      val_set_size_percentage: int, maf_percentage: int):
    """
    Check index file, described in create_index_file(), and append datasets if necessary

    :param data_dir: data directory where the phenotype and genotype matrix are stored
    :param genotype_matrix_name: name of the genotype matrix including datatype ending
    :param phenotype_matrix_name: name of the phenotype matrix including datatype ending
    :param phenotype: name of the phenotype to predict
    :param maf_percentage: threshold for MAF filter as percentage value
    :param datasplit: datasplit to use. Options are: nested-cv, cv-test, train-val-test
    :param n_outerfolds: number of outerfolds relevant for nested-cv
    :param n_innerfolds: number of folds relevant for nested-cv and cv-test
    :param test_set_size_percentage: size of the test set relevant for cv-test and train-val-test
    :param val_set_size_percentage: size of the validation set relevant for train-val-test
    """
    with h5py.File(data_dir.joinpath(genotype_matrix_name.split('.')[0] + '-'
                   + phenotype_matrix_name.split('.')[0] + '-' + phenotype + '.h5'), 'a') as f:
        # check if group 'maf_filter' is available and if user input maf is available, if not: create group/dataset
        if 'maf_filter' not in f:
            maf = f.create_group('maf_filter')
            tmp = (create_maf_filter(maf=maf_percentage, freq=f['matched_data/ma_frequency'][:]))
            maf.create_dataset(f'maf_{maf_percentage}', data=tmp, chunks=True, compression="gzip")
        elif f'maf_{maf_percentage}' not in f['maf_filter']:
            tmp = (create_maf_filter(maf=maf_percentage, freq=f['matched_data/ma_frequency'][:]))
            f.create_dataset(f'maf_filter/maf_{maf_percentage}', data=tmp, chunks=True, compression="gzip")
        # check if group datasplit and all user inputs concerning datasplits are available, if not: create all
        if datasplit == 'nested-cv':
            subpath = helper_functions.get_subpath_for_datasplit(datasplit=datasplit,
                                                                 datasplit_params=[n_outerfolds, n_innerfolds])
            if 'datasplits' not in f or ('datasplits' in f and 'nested-cv' not in f['datasplits']) or \
                    ('datasplits' in f and 'nested-cv' in f['datasplits'] and
                     f'{subpath}' not in f['datasplits/nested-cv']):
                nest = f.create_group(f'datasplits/nested-cv/{subpath}')
                index_dict = check_train_test_splits(y=f['matched_data/y'], datasplit='nested-cv',
                                                     datasplit_params=[n_outerfolds, n_innerfolds])
                for outer in range(n_outerfolds):
                    o = nest.create_group(f'outerfold_{outer}')
                    o.create_dataset('test', data=index_dict[f'outerfold_{outer}_test'], chunks=True,
                                     compression="gzip")
                    for inner in range(n_innerfolds):
                        i = o.create_group(f'innerfold_{inner}')
                        i.create_dataset('train', data=index_dict[f'outerfold_{outer}'][f'fold_{inner}_train'],
                                         chunks=True,
                                         compression="gzip")
                        i.create_dataset('val', data=index_dict[f'outerfold_{outer}'][f'fold_{inner}_test'],
                                         chunks=True,
                                         compression="gzip")
        elif datasplit == 'cv-test':
            subpath = helper_functions.get_subpath_for_datasplit(
                datasplit=datasplit, datasplit_params=[n_innerfolds, test_set_size_percentage]
            )
            if 'datasplits' not in f or ('datasplits' in f and 'cv-test' not in f['datasplits']) or \
                    ('datasplits' in f and 'cv-test' in f['datasplits'] and
                     f'{subpath}' not in f['datasplits/cv-test']):
                cv = f.create_group(f'datasplits/cv-test/{subpath}')
                index_dict, test = check_train_test_splits(y=f['matched_data/y'], datasplit='cv-test',
                                                           datasplit_params=[n_innerfolds, test_set_size_percentage])
                o = cv.create_group('outerfold_0')
                o.create_dataset('test', data=test, chunks=True, compression="gzip")
                for fold in range(n_innerfolds):
                    i = o.create_group(f'innerfold_{fold}')
                    i.create_dataset('train', data=index_dict[f'fold_{fold}_train'], chunks=True, compression="gzip")
                    i.create_dataset('val', data=index_dict[f'fold_{fold}_test'], chunks=True, compression="gzip")
        elif datasplit == 'train-val-test':
            subpath = helper_functions.get_subpath_for_datasplit(datasplit=datasplit,
                                                                 datasplit_params=[val_set_size_percentage,
                                                                                   test_set_size_percentage])
            if 'datasplits' not in f or ('datasplits' in f and 'train-val-test' not in f['datasplits']) or \
                    ('datasplits' in f and 'train-val-test' in f['datasplits'] and
                     f'{subpath}' not in f['datasplits/train-val-test']):
                tvt = f.create_group(f'datasplits/train-val-test/{subpath}')
                train, val, test = check_train_test_splits(y=f['matched_data/y'], datasplit='train-val-test',
                                                           datasplit_params=[val_set_size_percentage,
                                                                             test_set_size_percentage])
                o = tvt.create_group('outerfold_0')
                o.create_dataset('test', data=test, chunks=True, compression="gzip")
                i = o.create_group('innerfold_0')
                i.create_dataset('train', data=train, chunks=True, compression="gzip")
                i.create_dataset('val', data=val, chunks=True, compression="gzip")


def create_index_file(data_dir: pathlib.Path, genotype_matrix_name: str, phenotype_matrix_name: str, phenotype: str,
                      datasplit: str, n_outerfolds: int, n_innerfolds: int, test_set_size_percentage: int,
                      val_set_size_percentage: int, maf_percentage: int,
                      X: np.array, y: np.array,  sample_ids: np.array, X_index: np.array, y_index: np.array):
    """
    Create the .h5 index file containing the maf filters and data splits for the combination of genotype matrix,
    phenotype matrix and phenotype.
    It will be created using standard values additionally to user inputs for the maf filters and data splits.

    Unified format of .h5 file containing the maf filters and data splits:

    .. code-block:: python

            'matched_data': {
                    'y': matched phenotypic values,
                    'matched_sample_ids': sample ids of matched genotype/phenotype,
                    'X_index': indices of genotype matrix to redo matching,
                    'y_index': indices of phenotype vector to redo matching,
                    'ma_frequency': minor allele frequency of each SNP of genotype file to create new MAF filters
                    }
            'maf_filter': {
                    'maf_{maf_percentage}': indices of SNPs to delete  # (with MAF < maf_percentage),
                    ...
                    }
            'datasplits': {
                    'nested_cv': {
                            '#outerfolds-#innerfolds': {
                                    'outerfold_0': {
                                        'innerfold_0': {'train': indices_train, 'val': indices_val},
                                        ...
                                        'innerfold_n': {'train': indices_train, 'val': indices_val},
                                        'test': test_indices
                                        },
                                    ...
                                    'outerfold_m': {
                                        'innerfold_0': {'train': indices_train, 'val': indices_val},
                                        ...
                                        'innerfold_n': {'train': indices_train, 'val': indices_val},
                                        'test': test_indices
                                        }
                                    },
                            ...
                            }
                    'cv-test': {
                            '#folds-test_percentage': {
                                    'outerfold_0': {
                                        'innerfold_0': {'train': indices_train, 'val': indices_val},
                                        ...
                                        'innerfold_n': {'train': indices_train, 'val': indices_val},
                                        'test': test_indices
                                        }
                                    },
                            ...
                            }
                    'train-val-test': {
                            'train_percentage-val_percentage-test_percentage': {
                                    'outerfold_0': {
                                        'innerfold_0': {'train': indices_train, 'val': indices_val},
                                        'test': test_indices
                                        }
                                    },
                            ...
                            }
                    }

    Standard values for the maf filters and data splits:

    - maf thresholds: 1, 3, 5
    - folds (inner-/outerfolds for 'nested-cv' and folds for 'cv-test'): 5
    - test percentage (for 'cv-test' and 'train-val-test'): 20
    - val percentage (for 'train-val-test'): 20

    :param data_dir: data directory where the phenotype and genotype matrix are stored
    :param genotype_matrix_name: name of the genotype matrix including datatype ending
    :param phenotype_matrix_name: name of the phenotype matrix including datatype ending
    :param phenotype: name of the phenotype to predict
    :param maf_percentage: threshold for MAF filter as percentage value
    :param datasplit: datasplit to use. Options are: nested-cv, cv-test, train-val-test
    :param n_outerfolds: number of outerfolds relevant for nested-cv
    :param n_innerfolds: number of folds relevant for nested-cv and cv-test
    :param test_set_size_percentage: size of the test set relevant for cv-test and train-val-test
    :param val_set_size_percentage: size of the validation set relevant for train-val-test
    :param X: genotype in additive encoding to create ma-frequencies
    :param y: matched phenotype values
    :param sample_ids: matched sample ids of genotype/phenotype
    :param X_index: index file of genotype to redo matching
    :param y_index: index file of phenotype to redo matching
    """
    X, non_informative_filter = filter_non_informative_snps(X=X)
    freq = get_minor_allele_freq(X=X)
    maf_threshold = [0, 1, 3, 5]  # standard values for maf threshold
    if maf_percentage not in maf_threshold:  # add user input if needed
        maf_threshold.append(maf_percentage)
    param_nested = [[5, 5]]  # standard values for outer and inner folds for nested-cv
    param_cv = [[5, 20]]  # standard number of folds and test percentage for cv-test
    param_tvt = [[20, 20]]  # standard train and val percentages for train-val-test split
    param_nested = check_datasplit_user_input(user_datasplit=datasplit,
                                              user_n_outerfolds=n_outerfolds, user_n_innerfolds=n_innerfolds,
                                              user_test_set_size_percentage=test_set_size_percentage,
                                              user_val_set_size_percentage=val_set_size_percentage,
                                              datasplit='nested-cv', param_to_check=param_nested)
    param_cv = check_datasplit_user_input(user_datasplit=datasplit,
                                          user_n_outerfolds=n_outerfolds, user_n_innerfolds=n_innerfolds,
                                          user_test_set_size_percentage=test_set_size_percentage,
                                          user_val_set_size_percentage=val_set_size_percentage,
                                          datasplit='cv-test', param_to_check=param_cv)
    param_tvt = check_datasplit_user_input(user_datasplit=datasplit,
                                           user_n_outerfolds=n_outerfolds, user_n_innerfolds=n_innerfolds,
                                           user_test_set_size_percentage=test_set_size_percentage,
                                           user_val_set_size_percentage=val_set_size_percentage,
                                           datasplit='train-val-test', param_to_check=param_tvt)

    with h5py.File(data_dir.joinpath(genotype_matrix_name.split('.')[0] + '-'
                   + phenotype_matrix_name.split('.')[0] + '-' + phenotype + '.h5'), 'w') as f:
        # all data needed to redo matching of X and y and to create new mafs and new data splits
        data = f.create_group('matched_data')
        data.create_dataset('y', data=y, chunks=True, compression="gzip")
        data.create_dataset('matched_sample_ids', data=sample_ids.astype(bytes), chunks=True, compression="gzip")
        data.create_dataset('X_index', data=X_index, chunks=True, compression="gzip")
        data.create_dataset('y_index', data=y_index, chunks=True, compression="gzip")
        data.create_dataset('non_informative_filter', data=non_informative_filter, chunks=True, compression="gzip")
        data.create_dataset('ma_frequency', data=freq, chunks=True, compression="gzip")
        # create and save standard mafs and maf according to user input
        maf = f.create_group('maf_filter')
        for threshold in maf_threshold:
            tmp = (create_maf_filter(maf=threshold, freq=freq))
            maf.create_dataset(f'maf_{threshold}', data=tmp, chunks=True, compression="gzip")
        # create and save standard data splits and splits according to user input
        dsplit = f.create_group('datasplits')
        nest = dsplit.create_group('nested-cv')
        for elem in param_nested:
            n = nest.create_group(helper_functions.get_subpath_for_datasplit(datasplit='nested-cv',
                                                                             datasplit_params=elem))
            index_dict = check_train_test_splits(y=y, datasplit='nested-cv', datasplit_params=elem)
            for outer in range(elem[0]):
                o = n.create_group(f'outerfold_{outer}')
                o.create_dataset('test', data=index_dict[f'outerfold_{outer}_test'], chunks=True, compression="gzip")
                for inner in range(elem[1]):
                    i = o.create_group(f'innerfold_{inner}')
                    i.create_dataset('train', data=index_dict[f'outerfold_{outer}'][f'fold_{inner}_train'], chunks=True,
                                     compression="gzip")
                    i.create_dataset('val', data=index_dict[f'outerfold_{outer}'][f'fold_{inner}_test'], chunks=True,
                                     compression="gzip")
        cv = dsplit.create_group('cv-test')
        for elem in param_cv:
            index_dict, test = check_train_test_splits(y=y, datasplit='cv-test', datasplit_params=elem)
            n = cv.create_group(helper_functions.get_subpath_for_datasplit(datasplit='cv-test', datasplit_params=elem))
            o = n.create_group('outerfold_0')
            o.create_dataset('test', data=test, chunks=True, compression="gzip")
            for fold in range(elem[0]):
                i = o.create_group(f'innerfold_{fold}')
                i.create_dataset('train', data=index_dict[f'fold_{fold}_train'], chunks=True, compression="gzip")
                i.create_dataset('val', data=index_dict[f'fold_{fold}_test'], chunks=True, compression="gzip")
        tvt = dsplit.create_group('train-val-test')
        for elem in param_tvt:
            train, val, test = check_train_test_splits(y=y, datasplit='train-val-test', datasplit_params=elem)
            n = tvt.create_group(helper_functions.get_subpath_for_datasplit(datasplit='train-val-test',
                                                                            datasplit_params=elem))
            o = n.create_group('outerfold_0')
            o.create_dataset('test', data=test, chunks=True, compression="gzip")
            i = o.create_group('innerfold_0')
            i.create_dataset('train', data=train, chunks=True, compression="gzip")
            i.create_dataset('val', data=val, chunks=True, compression="gzip")


def filter_non_informative_snps(X: np.array) -> (np.array, np.array):
    """
    Remove non-informative SNPs, i.e. SNPs that are constant

    :param X: genotype matrix in raw or additive encoding

    :return: filtered genotype matrix and filter-vector
    """
    tmp = X == X[0, :]
    X_filtered = X[:, ~tmp.all(0)]
    return X_filtered, (~tmp.all(0)).nonzero()[0]


def get_minor_allele_freq(X: np.array) -> np.array:
    """
    Compute minor allele frequencies of genotype matrix

    :param X: genotype matrix in additive encoding

    :return: array with frequencies
    """
    encoding = enc.check_encoding_of_genotype(X=X)
    if encoding == '012':
        return (np.sum(X, 0)) / (2 * X.shape[0])
    else:
        freq = []
        for i, col in enumerate(np.transpose(X)):
            unique, counts = np.unique(col, return_counts=True)
            if len(unique) > 3:
                raise Exception('More than two alleles encountered at SNP ', i)
            elif len(unique) == 3:
                boolean = (unique.astype(str) == 'A') | (unique.astype(str) == 'T') | (unique.astype(str) == 'C') | (
                            unique.astype(str) == 'G')
                homozygous = unique[boolean]
                hetero = unique[~boolean][0]
                pairs = [['A', 'C'], ['A', 'G'], ['A', 'T'], ['C', 'G'], ['C', 'T'], ['G', 'T']]
                heterozygous_nuc = ['M', 'R', 'W', 'S', 'Y', 'K']
                for j, pair in enumerate(pairs):
                    if all(h in pair for h in homozygous) and hetero != heterozygous_nuc[j]:
                        raise Exception('More than two alleles encountered at SNP ' + str(i))
                freq.append((np.min(counts[boolean]) + 0.5 * counts[~boolean][0]) / len(col))
            else:
                freq.append(np.min(counts) / len(col))
        return np.array(freq)


def create_maf_filter(maf: int, freq: np.array) -> np.array:
    """
    Create minor allele frequency filter

    :param maf: maf threshold as percentage value
    :param freq: array containing minor allele frequencies as decimal value

    :return: array containing indices of SNPs with MAF smaller than specified threshold, i.e. SNPs to delete
    """
    return np.where(freq <= (maf / 100))[0]


def check_datasplit_user_input(user_datasplit: str, user_n_outerfolds: int, user_n_innerfolds: int,
                               user_test_set_size_percentage: int, user_val_set_size_percentage: int,
                               datasplit: str, param_to_check: list) -> list:
    """
    Check if user input of data split parameters differs from standard values.
    If it does, add input to list of parameters

    :param user_datasplit: datasplit specified by the user
    :param user_n_outerfolds: number of outerfolds relevant for nested-cv specified by the user
    :param user_n_innerfolds: number of folds relevant for nested-cv and cv-test specified by the user
    :param user_test_set_size_percentage: size of the test set relevant for cv-test and train-val-test specified by the user
    :param user_val_set_size_percentage: size of the validation set relevant for train-val-test specified by the user
    :param datasplit: type of data split
    :param param_to_check: standard parameters to compare to

    :return: adapted list of parameters
    """
    if datasplit == 'nested-cv':
        user_input = [user_n_outerfolds, user_n_innerfolds]
    elif datasplit == 'cv-test':
        user_input = [user_n_innerfolds, user_test_set_size_percentage]
    elif datasplit == 'train-val-test':
        user_input = [user_val_set_size_percentage, user_test_set_size_percentage]
    else:
        raise Exception('Only accept nested-cv, cv-test or train-val-test as data splits.')
    if user_datasplit == datasplit and user_input not in param_to_check:
        param_to_check.append(user_input)
    return param_to_check


def check_train_test_splits(y: np.array, datasplit: str, datasplit_params: list):
    """
    Create stratified train-test splits. Continuous values will be grouped into bins and stratified according to those.

    Datasplit parameters:

    - nested-cv: [n_outerfolds, n_innerfolds]
    - cv-test: [n_innerfolds, test_set_size_percentage]
    - train-val-test: [val_set_size_percentage, train_set_size_percentage]

    :param datasplit: type of datasplit ('nested-cv', 'cv-test', 'train-val-test')
    :param y: array with phenotypic values for stratification
    :param datasplit_params: parameters to use for split

    :return: dictionary respectively arrays with indices
    """
    y_binned = make_bins(y=y, datasplit=datasplit, datasplit_params=datasplit_params)
    if datasplit == 'nested-cv':
        return make_nested_cv(y=y_binned, outerfolds=datasplit_params[0], innerfolds=datasplit_params[1])
    elif datasplit == 'cv-test':
        x_train, x_test, y_train = make_train_test_split(y=y_binned, test_size=datasplit_params[1], val=False)
        cv_dict = make_stratified_cv(x=x_train, y=y_train, split_number=datasplit_params[0])
        return cv_dict, x_test
    elif datasplit == 'train-val-test':
        return make_train_test_split(y=y_binned, test_size=datasplit_params[1], val_size=datasplit_params[0], val=True)
    else:
        raise Exception('Only accept nested-cv, cv-test or train-val-test as data splits.')


def make_bins(y: np.array, datasplit: str, datasplit_params: list) -> np.array:
    """
    Create bins of continuous values for stratification.

    Datasplit parameters:

    - nested-cv: [n_outerfolds, n_innerfolds]
    - cv-test: [n_innerfolds, test_set_size_percentage]
    - train-val-test: [val_set_size_percentage, train_set_size_percentage]

    :param y: array containing phenotypic values
    :param datasplit: train test split to use
    :param datasplit_params: parameters to use for split

    :return: binned array
    """
    if helper_functions.test_likely_categorical(y):
        return y.astype(int)
    else:
        if datasplit == 'nested-cv':
            tmp = len(y)/(datasplit_params[0] + datasplit_params[1])
        elif datasplit == 'cv-test':
            tmp = len(y)*(1-datasplit_params[1]/100)/datasplit_params[0]
        else:
            tmp = len(y)/10 + 1

        number_of_bins = min(int(tmp) - 1, 10)
        edges = np.percentile(y, np.linspace(0, 100, number_of_bins)[1:])
        y_binned = np.digitize(y, edges, right=True)
        return y_binned


def make_nested_cv(y: np.array, outerfolds: int, innerfolds: int) -> dict:
    """
    Create index dictionary for stratified nested cross validation with the following structure:

    .. code-block:: python

        {
            'outerfold_0_test': test_indices,
            'outerfold_0': {
                'fold_0_train': innerfold_0_train_indices,
                'fold_0_test': innerfold_0_test_indices,
                ...
                'fold_n_train': innerfold_n_train_indices,
                'fold_n_test': innerfold_n_test_indices
            },
            ...
            'outerfold_m_test': test_indices,
            'outerfold_m': {
                'fold_0_train': innerfold_0_train_indices,
                'fold_0_test': innerfold_0_test_indices,
                ...
                'fold_n_train': innerfold_n_train_indices,
                'fold_n_test': innerfold_n_test_indices
            }
        }

    :param y: target values grouped in bins for stratification
    :param outerfolds: number of outer folds
    :param innerfolds: number of inner folds

    :return: index dictionary
    """
    outer_cv = StratifiedKFold(n_splits=outerfolds, shuffle=True)
    index_dict = {}
    outer_fold = 0
    for train_index, test_index in outer_cv.split(np.zeros(len(y)), y):
        np.random.shuffle(test_index)
        index_dict[f'outerfold_{outer_fold}_test'] = test_index
        index_dict[f'outerfold_{outer_fold}'] = make_stratified_cv(x=train_index, y=y[train_index],
                                                                   split_number=innerfolds)
        outer_fold += 1
    return index_dict


def make_stratified_cv(x: np.array, y: np.array, split_number: int) -> dict:
    """
    Create index dictionary for stratified cross-validation with following structure:

    .. code-block:: python

        {
            'fold_0_train': fold_0_train_indices,
            'fold_0_test': fold_0_test_indices,
            ...
            'fold_n_train': fold_n_train_indices,
            'fold_n_test': fold_n_test_indices
        }

    :param x: whole train indices
    :param y: target values binned in groups for stratification
    :param split_number: number of folds

    :return: dictionary containing train and validation indices for each fold
    """
    cv = StratifiedKFold(n_splits=split_number, shuffle=True)
    index_dict = {}
    fold = 0
    for train_index, test_index in cv.split(x, y):
        np.random.shuffle(train_index)
        np.random.shuffle(test_index)
        index_dict[f'fold_{fold}_train'] = x[train_index]
        index_dict[f'fold_{fold}_test'] = x[test_index]
        fold += 1
    return index_dict


def make_train_test_split(y: np.array, test_size: int, val_size=None, val=False, random=42) \
        -> (np.array, np.array, np.array):
    """
    Create index arrays for stratified train-test, respectively train-val-test splits.

    :param y: target values grouped in bins for stratification
    :param test_size: size of test set as percentage value
    :param val_size: size of validation set as percentage value
    :param val: if True, function returns validation set additionally to train and test set
    :param random: controls shuffling of data

    :return: either train, val and test index arrays or train and test index arrays and corresponding binned target values
    """
    x = np.arange(len(y))
    x_train, x_test, y_train, y_test = train_test_split(x, y, test_size=test_size/100, stratify=y, random_state=random)
    if not val:
        return x_train, x_test, y_train
    else:
        x_train, x_val, y_train, y_val = train_test_split(x_train, y_train, test_size=val_size/100, stratify=y_train,
                                                          random_state=random)
        return x_train, x_val, x_test



__all__ = [
    "_base_model", "_param_free_base_model", "_model_functions", "_sklearn_model", "_tensorflow_model", "_torch_model",
    "_bayesian_linreg", "_template_torch_model", "_template_sklearn_model", "_template_tensorflow_model",
    "cnn", "linearregression", "elasticnet", "localcnn", "mlp", "randomforest", "svm", "xgboost", "blup", "bayes_ridge"]

from . import *



import argparse

from . import model_reuse

if __name__ == "__main__":
    """
    Run to apply the specified model on a dataset containing new samples.
    """
    parser = argparse.ArgumentParser()
    parser.add_argument("-odd", "--old_data_dir", type=str, default=None,
                        help="Provide the full path of the old data directory (that contains the geno- and phenotype "
                             "files as well as the index file the model was trained on).")
    parser.add_argument("-ndd", "--new_data_dir", type=str,
                        help="Provide the full path of the new data directory that contains the geno- and phenotype "
                             "files you want to predict on")
    parser.add_argument("-ngm", "--new_genotype_matrix", type=str,
                        help="Provide the name of the new genotype matrix you want to predict on")
    parser.add_argument("-npm", "--new_phenotype_matrix", type=str,
                        help="Provide the name of the new phenotype matrix you want to predict on")
    parser.add_argument("-sd", "--save_dir", type=str,
                        help="Define the save directory for the results.")
    parser.add_argument("-rd", "--results_dir_model", type=str,
                        help="Provide the full path of the directory where your results of the model "
                             "you want to use are stored")
    args = vars(parser.parse_args())
    old_data_dir = args['old_data_dir']
    new_data_dir = args['new_data_dir']
    new_genotype_matrix = args['new_genotype_matrix']
    new_phenotype_matrix = args['new_phenotype_matrix']
    save_dir = args["save_dir"]
    results_directory_model = args['results_dir_model']

    model_reuse.apply_final_model(
        results_directory_model=results_directory_model, old_data_dir=old_data_dir, new_data_dir=new_data_dir,
        new_genotype_matrix=new_genotype_matrix, new_phenotype_matrix=new_phenotype_matrix, save_dir=save_dir
    )



import numpy as np
from scipy.optimize import minimize
from . import _param_free_base_model


class Blup(_param_free_base_model.ParamFreeBaseModel):
    """
    Implementation of a class for BLUP.

    *Attributes*

        *Inherited attributes*

        See :obj:`~easypheno.model._param_free_base_model.ParamFreeBaseModel` for more information on the attributes.

        *Additional attributes*

        - beta (*np.array*): best linear unbiased estimate (BLUE) of the fixed effects
        - u (*np.array*): best linear unbiased prediction (BLUP) of the random effects
    """
    standard_encoding = '101'
    possible_encodings = ['101']

    def __init__(self, task: str, encoding: str = None):
        super().__init__(task=task, encoding=encoding)
        self.beta = None
        self.u = None

    @staticmethod
    def reml(delta: float, n: int, eigenvalues: np.array, omega2: np.array):
        """
        Function to compute the restricted maximum likelihood

        :param delta: variance component
        :param n: number of samples
        :param eigenvalues: eigenvalues of SHS
        :param omega2: point-wise product of V_SHS*y with itself
        :return: restricted maximum likelihood
        """
        return (n-1)*np.log(np.sum(np.divide(omega2, (eigenvalues+delta)))) + np.sum(np.log(eigenvalues+delta))

    def fit(self, X: np.array, y: np.array) -> np.array:
        """
        Implementation of fit function for BLUP.

        See :obj:`~easypheno.model._param_free_base_model.ParamFreeBaseModel` for more information.
        """
        n = y.shape[0]
        Z = np.ones((n, 1))
        S = np.eye(n) - 1/n*np.ones((n, n))
        sqn = np.sqrt(n)

        # compute spectral decomposition
        H = np.matmul(X, X.T) + sqn*np.eye(n)
        eigenvalues_H, V_H = np.linalg.eigh(H)
        eigenvalues_H = eigenvalues_H - sqn
        SHS = np.matmul(S, np.matmul(H, S))
        eigenvalues_SHS, V_SHS = np.linalg.eigh(SHS)
        V_SHS = np.delete(V_SHS, -1, axis=1)
        eigenvalues_SHS = np.delete(eigenvalues_SHS, -1) - sqn
        omega = np.matmul(V_SHS.T, y)
        omega2 = np.multiply(omega, omega)

        # minimize REML
        x0 = np.abs(np.min(eigenvalues_SHS)) + 1e-09
        delta_opt = minimize(self.reml, x0=x0, args=(n, eigenvalues_SHS, omega2), bounds=[(1e-09, 1e+09)]).x

        # calculate inverse of H
        H_inv = np.matmul(V_H, np.divide(V_H, (eigenvalues_H + delta_opt)).T)

        # calculate beta and u
        self.beta = np.linalg.solve(np.matmul(Z.T, np.matmul(H_inv, Z)), np.matmul(Z.T, np.matmul(H_inv, y)))
        self.u = np.matmul(X.T, np.matmul(H_inv, (y - np.matmul(Z, self.beta))))
        return self.predict(X_in=X)

    def predict(self, X_in: np.array) -> np.array:
        """
        Implementation of predict function for BLUP.

        See :obj:`~easypheno.model._param_free_base_model.ParamFreeBaseModel` for more information.
        """
        return self.beta + np.matmul(X_in, self.u)
